/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              SCOPING OF CONDENSATIONS                  * */
/* *                                                        * */
/* *  $Module: condensationscope                            * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _CONDENSATION_
#define _CONDENSATION_


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "evector.h"
#include "position.h"


/**************************************************************/
/* Macros                                                     */
/**************************************************************/

#define condensation_scope_mmm_malloc(size) memory_Malloc(size)
#define condensation_scope_mmm_free(ptr, size) memory_Free(ptr)


/**************************************************************/
/* Datastructures                                             */
/**************************************************************/

/* stores position of condensed _SPLITS_ */
typedef EVECTOR CONDENSATION_SCOPE;


/**************************************************************/
/* Creation and deletion                                      */
/**************************************************************/

CONDENSATION_SCOPE* condensationscope_createCONDENSATIONSCOPE(int);
void condensationscope_destroyCONDENSATIONSCOPE(CONDENSATION_SCOPE*);


/**************************************************************/
/* Adding and retrieval                                       */
/**************************************************************/

void condensationscope_addCondensation(CONDENSATION_SCOPE*, POSITION*);
int condensationscope_getNumberOfCondensationsUntil(CONDENSATION_SCOPE*, POSITION*);
int condensationscope_dependsOnCondensedBranch(CONDENSATION_SCOPE*, EVECTOR*);
int condensationscope_wasCondensedBranch(CONDENSATION_SCOPE*, POSITION *);

static __inline__
int condensationscope_getTotalNumberOfCondensations(CONDENSATION_SCOPE *scope)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object.
  RETURNS: How many branches have been condensed so far.
***************************************************************/
{
  return(scope->number_of_elements);
}

static __inline__
int condensationscope_wasCondensedSplit(CONDENSATION_SCOPE *scope, POSITION *p)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object, a position of a split.
  RETURNS: 1 if <p> was condensed, 0 otherwise.
***************************************************************/
{
  return (evector_isMember(scope, (void*) p));
}

#endif
