/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   EXPANDABLE STRINGS                   * */
/* *                                                        * */
/* *  $Module: estrings                                     * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _ESTRINGS_
#define _ESTRINGS_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <math.h>
#include "memory.h"


/**************************************************************/
/* Data Structures                                            */
/**************************************************************/

/* Datastructure invariant: The <string> field was allocated as a block
   of <current_size>*sizeof(char) bytes.
*/

typedef struct ESTRING_HELP {
  int current_size;
  char *string;
} ESTRING;


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

BOOL      estring_Equal(const ESTRING*, const ESTRING*);
BOOL      estring_EqualString(const ESTRING*, const char*);
ESTRING*  estring_EmptyString(int);
ESTRING*  estring_FromString(const char*);
void      estring_StringFree(ESTRING*);
void      estring_AppendString(ESTRING*, const char*);
void      estring_AppendEString(ESTRING*, const ESTRING*);
void      estring_AppendStringDelete(ESTRING*, char*);
void      estring_AppendStringDeleteMPN(ESTRING*, char*);
void      estring_AppendEStringDelete(ESTRING*, ESTRING*);
char*     estring_ExtractString(ESTRING* estring);


/**************************************************************/
/* Inline Functions                                           */
/**************************************************************/


static __inline__ char* estring_GetString(const ESTRING* estring)
{
  return estring->string;
}

static __inline__ int estring_GetCurrentSize(const ESTRING* estring)
{
  return estring->current_size;
}

static __inline__ void estring_SetString(ESTRING* estring, char *string)
{
  estring->string = string;
}

static __inline__ void estring_SetCurrentSize(ESTRING* estring, int new_size)
{
  estring->current_size = new_size;
}


#ifdef __cplusplus
}
#endif

#endif
