/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                CONNECTION TO CVC3                      * */
/* *                                                        * */
/* *  $Module: smt_cvc3                                     * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/*                                                            */
/*    Notes regarding the external SMT procedure "cvc3"       */
/*                                                            */
/*                                                            */
/* Note that in the following the term "external SMT          */
/* procedure" refers to the prover "cvc3".                    */
/* The present version of SPASS+T supports cvc3 in the        */
/* versions 1.0 and 1.2.1. CVC3 can be obtained from          */
/*    http://www.cs.nyu.edu/acsys/cvc3/download.html          */
/*                                                            */
/* For any details including license terms look for           */
/*    http://www.cs.nyu.edu/acsys/cvc3/index.html             */
/*                                                            */
/*                                                            */
/**************************************************************/

#ifndef _SMTCVC3_
#define _SMTCVC3_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dpconnect.h"
#include "smt_interface.h"


/**************************************************************/
/* Function prototypes                                        */
/**************************************************************/

SMT_INTERFACE* smt_cvc3_getInterface(void);

void smt_cvc3_Init(void);
void smt_cvc3_Free(void);

EVECTOR* smt_cvc3_getCommandLineArguments(struct PROOFSEARCH_HELP*);
int smt_cvc3_parseSMTOutput(struct PROOFSEARCH_HELP*, char*, EVECTOR**, EVECTOR**, int);
void smt_cvc3_printAssertion(struct PROOFSEARCH_HELP*, FILE*, char*, int, int);
void smt_cvc3_printCheck(struct PROOFSEARCH_HELP*, FILE*);
void smt_cvc3_printSPASSSignature(struct PROOFSEARCH_HELP*, FILE*);
void smt_cvc3_printComment(struct PROOFSEARCH_HELP*, FILE*,char*);
void smt_cvc3_prepareNonFilteredClauseInput(struct PROOFSEARCH_HELP*, FILE*);

/**************************************************************/
/* Conversion Functions                                       */
/**************************************************************/

ESTRING* smt_cvc3_Cvc3QuantifierPrefixToString(CLAUSE, char*);
ESTRING* smt_cvc3_Cvc3ClauseToString(CLAUSE, char*);
ESTRING* smt_cvc3_Cvc3AtomToString(TERM, BOOL);
ESTRING* smt_cvc3_Cvc3TermToString(TERM, BOOL);
ESTRING* smt_cvc3_Cvc3TermListToString(LIST, BOOL);
void     smt_cvc3_checkIdentifier(char*);
ESTRING* smt_cvc3_Cvc3SigToString(const char*);
ESTRING* smt_cvc3_Cvc3NumericSymbolToString(SYMBOL);

ESTRING* smt_cvc3_Cvc3FormulaToString(TERM, char*, BOOL);

void smt_cvc3_FPrintCnfCvc3(FILE*, LIST, char*, int);
void smt_cvc3_FPrintCvc3(FILE*, LIST, FLAG_TDFG2OTTEROPTIONSTYPE, char*, int);

#endif
