/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               CONNECTION TO YICES                      * */
/* *                                                        * */
/* *  $Module: smt_yices                                    * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/*                                                            */
/*    Notes regarding the external SMT procedure "yices"      */
/*                                                            */
/*                                                            */
/* Note that in the following the term "external SMT          */
/* procedure" refers to the prover "yices".                   */
/* The present version of SPASS+T supports yices in the       */
/* version 1.0.x; Yices can be obtained from                  */
/*     http://yices.csl.sri.com/download.shtml                */
/*                                                            */
/* For any details including license terms look for           */
/*     http://yices.csl.sri.com/index.shtml                   */
/*                                                            */
/*                                                            */
/**************************************************************/

#ifndef _SMTYICES_
#define _SMTYICES_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dpconnect.h"
#include "smt_interface.h"


/**************************************************************/
/* Function prototypes                                        */
/**************************************************************/

SMT_INTERFACE* smt_yices_getInterface(void);

void smt_yices_Init(void);
void smt_yices_Free(void);

EVECTOR* smt_yices_getCommandLineArguments(struct PROOFSEARCH_HELP*);
int smt_yices_parseSMTOutput(struct PROOFSEARCH_HELP*, char*, EVECTOR**, EVECTOR**, int);
void smt_yices_printAssertion(struct PROOFSEARCH_HELP*, FILE*, char*, int, int);
void smt_yices_printCheck(struct PROOFSEARCH_HELP*, FILE*);
void smt_yices_printSPASSSignature(struct PROOFSEARCH_HELP*, FILE*);
void smt_yices_printComment(struct PROOFSEARCH_HELP*, FILE*,char*);
void smt_yices_prepareNonFilteredClauseInput(struct PROOFSEARCH_HELP*, FILE*);

/**************************************************************/
/* Conversion Functions                                       */
/**************************************************************/

void smt_yices_YicesClausePrint(FILE*, CLAUSE);
void smt_yices_YicesAtomPrint(FILE*, TERM);
void smt_yices_YicesTermPrint(FILE*, TERM);
void smt_yices_YicesTermListPrint(FILE*, LIST);
void smt_yices_YicesPrintSig(FILE*, const char*);

ESTRING* smt_yices_YicesQuantifierPrefixToString(CLAUSE, char*);
ESTRING* smt_yices_YicesClauseToString(CLAUSE, char*);

ESTRING* smt_yices_YicesFormulaToString(TERM, char*, BOOL);

ESTRING* smt_yices_YicesAtomToString(TERM, BOOL);
ESTRING* smt_yices_YicesTermToString(TERM, BOOL);
ESTRING* smt_yices_YicesTermListToString(LIST, BOOL);
void     smt_yices_checkIdentifier(char*);
ESTRING* smt_yices_YicesSigToString(const char*);
ESTRING* smt_yices_YicesNumericSymbolToString(SYMBOL);

void smt_yices_FPrintCnfYices(FILE*, LIST, char*, int);
void smt_yices_FPrintYices(FILE*, LIST, FLAG_TDFG2OTTEROPTIONSTYPE, char*, int);

#endif
