/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *      Connection to GNU Multiprecision Library          * */
/* *                                                        * */
/* *  $Module:   SPASSGMP                                   * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: July 2007                                         * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _SPASS_GMP_
#define _SPASS_GMP_


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <gmp.h>
#include "misc.h"


/**************************************************************/
/* Data Structures, Constants, and Type Definitions           */
/**************************************************************/

/* In our setting MULTIPRECESION NUMBERS are GMP rational numbers;
   they represent both,
        * integers (possibly bignums) and
	* rational numbers
   We must maintain the following invariants:
        * if a symbol denotes a number, its representation wich
          is stored in SIGNATURE_NODE must be canonicalzed, i.e.
          <mpq_canonicalize> must have been called before
        * whenever a symbol denotes an integer (i.e. the flag
          DENOTESINT is set), the denominator of the denoted
          number must be equal to 1 (this might be subject to further
          changes: maybe we drop the flag since we're not aiming to use
          sorts anymore anyway, so we do not need this seperation)
*/
typedef mpq_t MPN;

#ifdef SPASS_GMP_DEBUG
/* how big is the memory overhead created by using a GMP libary */
extern long gmp_memory_allocated_total;
extern long gmp_memory_used_balance;
extern long gmp_memory_used_pike;
#endif


/**************************************************************/
/* Function Declarations                                      */
/**************************************************************/

void gmp_Init(void);
void gmp_Free(void);

MPN* gmp_AllocMPN(void);
MPN* gmp_CreateMPN(signed long int, unsigned long int);
void gmp_FreeMPN(MPN*);
char* gmp_MPNToName(MPN);
MPN* gmp_NameToMPN(const char*);

BOOL gmp_IsMPNRational(MPN r);
BOOL gmp_TryMPNToInt(MPN, int *);

BOOL gmp_StringIsMPN(char*);

mpz_t* gmp_RoundUpwards(MPN);

#endif
