/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              TOP MODULE OF SPASS                       * */
/* *                                                        * */
/* *  $Module:   TOP                                        * */ 
/* *                                                        * */
/* *  Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001      * */
/* *  MPI fuer Informatik                                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.120 $                                       * */
/* $State: Exp $                                            * */
/* $Date: 2005/10/04 13:01:09 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-inf.mpg.de              * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: top.c,v $ */

/*** MAIN LOOP *************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfg.h"
#include "defs.h"
#include "ia.h"
#include "rules-inf.h"
#include "rules-sort.h"
#include "rules-split.h"
#include "terminator.h"
#include "rules-red.h"
#include "analyze.h"
#include "clock.h"
#include "strings.h"
#include "options.h"
#include "context.h"
#include "cnf.h"
#include "search.h"
#include "hasharray.h"
#include "closure.h"
/* (UW) */
#include "dpconnect.h"
#include <errno.h>
#include <stdlib.h>

/* SZ > */
/**************************************************************/
/* Debug Switches                                             */
/**************************************************************/

/* #define __DEBUG_SPLITTING__  */
/* < SZ */


/**************************************************************/
/* Types and Variables                                        */
/**************************************************************/

static const char *top_InputFile;

static OPTID top_RemoveFileOptId;

typedef enum {top_PROOF, 
	      top_COMPLETION, 
	      top_RESOURCE
	      /* SZ > */,top_SMT, top_SPASS_SMT, top_SMT_SAT /* < SZ */
} top_SEARCHRESULT;

/* SZ > */
/* the linearnumber of the first occurrence of a DFG_FRAC symbol */
NAT     top_FIRSTFRAC=0;
/* < SZ */

/**************************************************************/
/* Catch Signals Section                                      */
/**************************************************************/

#ifdef SPASS_SIGNALS
#include <signal.h>

static PROOFSEARCH *top_PROOFSEARCH;

static void top_SigHandler(int Signal)
/**************************************************************
  INPUT:   
  RETURNS: 
  EFFECT:  
***************************************************************/
{
  /* SZ > */
  if (prfs_DpConnect(*top_PROOFSEARCH) != (DPCONNECT)NULL) {
    puts("\nKilling decision procedure subprocess.");
    dpconnect_interruptExternalSMT(*top_PROOFSEARCH, NULL, "top_SigHandler", TRUE);
    /* memory for "dpconnect" structure is freed later */
  }
  /* < SZ */

  if (Signal == SIGSEGV || Signal == SIGBUS) {
    if (Signal == SIGSEGV) {
      puts("\n\nSEGMENTATION FAULT\n");
    }

    puts("\n\n\tSPASS is going to crash. This is probably caused by a");
    puts("\tbug in SPASS.  Please send a copy of the input file(s) together");
    puts("\twith the used options to weidenb@mpi-inf.mpg.de, and someone will");
    puts("\ttry to fix the problem.\n");
    abort();
  }

  if (Signal == SIGINT || Signal == SIGTERM) {
    clock_StopPassedTime(clock_OVERALL);
    printf("\n\n%% SZS status Forced for %s\n",
           (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
    printf("\nSPASS %s ", misc_VERSION);
    puts("\nSPASS beiseite: Ran out of time. SPASS was killed.");
    printf("Problem: %s ", 
	   (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
    printf("\nSPASS derived %d clauses, backtracked %d clauses "
	   "and kept %d clauses.",
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_DerivedClauses(*top_PROOFSEARCH)),
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_BacktrackedClauses(*top_PROOFSEARCH)),
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_KeptClauses(*top_PROOFSEARCH)));
    printf("\nSPASS allocated %lu KBytes.", memory_DemandedBytes()/1024);
    fputs("\nSPASS spent\t", stdout);
    clock_PrintTime(clock_OVERALL);
    fputs(" on the problem.\n\t\t", stdout);
    clock_PrintTime(clock_INPUT);
    fputs(" for the input.\n\t\t", stdout);
    clock_PrintTime(clock_CNF);
    fputs(" for the FLOTTER CNF translation.\n\t\t", stdout);
    clock_PrintTime(clock_INFERENCE);
    fputs(" for inferences.\n\t\t", stdout);
    clock_PrintTime(clock_BACKTRACK);
    fputs(" for the backtracking.\n\t\t", stdout);
    clock_PrintTime(clock_REDUCTION);
    puts(" for the reduction.");
    /* SZ > */
    if (flag_GetFlagValue(prfs_Store(*top_PROOFSEARCH),flag_DPCPROVER) != flag_DPCOFF) {
      fputs("\t\t", stdout);
      clock_PrintTime(clock_SMT);
      fputs(" for interacting with the SMT procedure.\n\t\t", stdout);
#ifdef SMT_DEBUG
      fputs("\n\t\t", stdout);
      clock_PrintTime(clock_SMTCREATEINPUT);
      fputs(" for providing input to the SMT procedure.", stdout);
      
      fputs("\n\t\t", stdout);
      clock_PrintTime(clock_SMTGETRESULT);
      fputs(" for {looking for,processing} the SMT result.\n", stdout);
#endif
    }
    /* < SZ */
  }

  if (opts_IsSet(top_RemoveFileOptId))
    remove(top_InputFile);

  exit(EXIT_FAILURE);
}

#endif


/**************************************************************/
/* Clause Selection Functions                                 */
/**************************************************************/

static CLAUSE top_SelectClauseDepth(LIST List, FLAGSTORE Flags)
/**************************************************************
  INPUT:   A list of clauses and a flag store.
  RETURNS: A clause selected from the list.
  EFFECT:  This function selects a clause from the list according
           to the following criteria:
	   1. minimal depth
           2. minimal weight
	   3a. maximal number of variable occurrences, if the flag
	       'PrefVar' is TRUE
	   3b. minimal number of variable occurrences, if 'PrefVar'
	       is FALSE
***************************************************************/
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight,Depth,NewDepth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewDepth =  clause_Depth(list_Car(List));
    if (NewDepth <= Depth) {
      if (NewDepth < Depth || clause_Weight(list_Car(List)) < Weight) {
	Depth  = NewDepth;
	Result = (CLAUSE)list_Car(List);
	Weight = clause_Weight(Result);
	Vars   = clause_NumberOfVarOccs(list_Car(List));
      }
      else
	if (clause_Weight(list_Car(List)) == Weight) {
	  NewVars = clause_NumberOfVarOccs(list_Car(List));
	  if (flag_GetFlagValue(Flags, flag_PREFVAR)) {
	    if (Vars < NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	  }
	  else
	    if (Vars > NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	}
    }
    List = list_Cdr(List);
  }

  return Result;
}


static CLAUSE top_SelectMinimalWeightClause(LIST List, FLAGSTORE Flags)
/**************************************************************
  INPUT:   A list of clauses and a flag store.
  RETURNS: A clause selected from the list.
  EFFECT:  This function selects a clause with minimal weight.
           If more than one clause has minimal weight and the flag
	   'PrefVar' is TRUE, a clause with maximal number of variable
	   occurrences is selected. If 'PrefVar' is FALSE, a clause with
	   minimal number of variable occurrences is selected.
	   If two clauses are equal with respect to the two criteria
	   the clause with the smaller list position is selected.
  CAUTION: THE LIST HAS TO BY SORTED BY WEIGHT IN ASCENDING ORDER!
***************************************************************/
{
  CLAUSE Result;
  NAT    Vars, NewVars, Weight;

#ifdef CHECK
  /* Check invariant: List has to be sorted by weight (ascending) */
  LIST Scan;
  Weight = clause_Weight(list_Car(List));
  for (Scan = list_Cdr(List); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    NAT NewWeight;
    NewWeight = clause_Weight(list_Car(Scan));
    if (NewWeight < Weight) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In top_SelectMinimalConWeightClause: clause list ");
      misc_ErrorReport("isn't sorted by weight");
      misc_FinishErrorReport();
    }
    Weight = NewWeight;
  }
#endif

  Result = (CLAUSE)list_Car(List);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    if (clause_Weight(list_Car(List)) == Weight) {
      NewVars = clause_NumberOfVarOccs(list_Car(List));
      if (flag_GetFlagValue(Flags, flag_PREFVAR)) {
	if (Vars < NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
      }
      else
	if (Vars > NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
    }
    else
      return Result;
    List = list_Cdr(List);
  }
  return Result;
}


static CLAUSE top_SelectMinimalConWeightClause(LIST List, FLAGSTORE Flags)
/**************************************************************
  INPUT:   A non-empty list of clauses and a flag store.
  RETURNS: A clause selected from the list.
  EFFECT:  This function selects a clause from the list in a
           similar way as the function top_SelectMinimalWeightClause.
	   The only difference is that conjecture clauses are
	   preferred over axiom clauses, because their weight
	   is divided by a factor given by the flag 'PrefCon'.
***************************************************************/
{
  CLAUSE Result;
  NAT    NewWeight,Weight, NewVars, Vars, Factor;

  Result = (CLAUSE)list_Car(List);
  Factor = flag_GetFlagValue(Flags, flag_PREFCON);
  Weight = clause_Weight(Result);
  if (clause_GetFlag(Result, CONCLAUSE))
    Weight = Weight / Factor;
  Vars   = clause_NumberOfVarOccs(list_Car(List));
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewWeight = clause_Weight(list_Car(List));
    if (clause_GetFlag(list_Car(List),CONCLAUSE))
      NewWeight = NewWeight / Factor;
    if (NewWeight < Weight) {
      Weight = NewWeight;
      Result = list_Car(List);
      Vars   = clause_NumberOfVarOccs(list_Car(List));
    }
    else {
      if (NewWeight == Weight) {
	NewVars = clause_NumberOfVarOccs(list_Car(List));
	if (flag_GetFlagValue(Flags, flag_PREFVAR)) {
	  if (Vars < NewVars) {
	    Result = (CLAUSE)list_Car(List);
	    Weight = NewWeight;
	    Vars   = NewVars;
	  }
	}
	else
	  if (Vars > NewVars) {
	    Result = (CLAUSE)list_Car(List);
	    Weight = NewWeight;
	    Vars   = NewVars;
	  }
      }
    }

    List = list_Cdr(List);
  }
  return Result;
}


/*static CLAUSE top_SelectClauseDepth(LIST List)*/
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: A clause selected from the list.
  EFFECT:  
***************************************************************/
/*{
  CLAUSE Result;
  int    Min, Depth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Min    = Depth * clause_Weight(Result);
  List   = list_Cdr(List);

  if (Depth == 0)
    return Result;

  while (!list_Empty(List)) {
    Depth =  clause_Depth(list_Car(List));
    if (Min > Depth * clause_Weight(list_Car(List))) {
      Result = list_Car(List);
      if (Depth == 0)
	return Result;
      Min    = clause_Depth(Result) * clause_Weight(Result);
    }
    List = list_Cdr(List);
  }

  return Result;
}*/

#define EQUAL(s,t)        term_Create(fol_Equality(),list_Cons((s),list_List(t)))
#define EVALEQ(s,t)       term_Create(fol_ThEvaleq(),list_Cons((s),list_List(t)))
#define LESS(s,t)         term_Create(fol_ThLess(),list_Cons((s),list_List(t)))
#define LESSEQ(s,t)       term_Create(fol_ThLesseq(),list_Cons((s),list_List(t)))
#define ISINT(s)          term_Create(fol_ThIsInt(),list_List(s))
#define ISRAT(s)          term_Create(fol_ThIsRat(),list_List(s))
#define PLUS(s,t)         term_Create(fol_ThPlus(),list_Cons((s),list_List(t)))
#define TIMES(s,t)        term_Create(fol_ThTimes(),list_Cons((s),list_List(t)))
#define DIVIDENAT(s,t)    term_Create(fol_ThDivideNat(),list_Cons((s),list_List(t)))
#define REMAINDER(s,t)    term_Create(fol_ThRemainder(),list_Cons((s),list_List(t)))
#define UMINUS(s)         term_Create(fol_ThUminus(),list_List(s))
#define FLOOR(s)          term_Create(fol_ThFloor(),list_List(s))
#define CEILING(s)        term_Create(fol_ThCeiling(),list_List(s))
#define ZERO              term_Create(fol_ThZero(),list_Nil())
#define ONE               term_Create(fol_ThOne(),list_Nil())
#define MINUSONE          term_Create(fol_ThMinusOne(),list_Nil())
#define cX                term_Copy(X)
#define cY                term_Copy(Y)
#define cZ                term_Copy(Z)

/* SZ > */
void top_AddStandardAxioms(PROOFSEARCH Search, LIST *ProblemClauses, int TheoryFlag)
/**************************************************************
  INPUT:   A ProofSearch object, a list of clauses, and
           a flag value from [flag_DPCTINT ... flag_DPCTREAL]
  EFFECT:  Adds the standard axioms of SPASS+T to the list 
           of <ProblemClauses>. Memory is allocated for the 
           new clauses. General properties of the newly 
           introduced clauses (weight, orientation of 
           equalities, and maximal variable) are also set up.
***************************************************************/
{
  CLAUSE NewClause;
  TERM X, Y, Z, NewAtom;
  LIST StandardAxioms, ExtensionAxioms, Scan;
  /* The only difference between StandardAxioms and ExtensionAxioms is
     that the latter are exported to the SMT procedure */
  
  /* initialization */
  StandardAxioms = list_Nil();
  ExtensionAxioms = list_Nil();

  /*** NEW formula(forall([X],equal(uminus(uminus(X)),X))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(UMINUS(UMINUS(X)),cX);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW2 formula(forall([X,Y],equal(plus(Y,plus(uminus(Y),X)),X))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(Y,PLUS(UMINUS(cY),X)),cX);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW2 formula(forall([X,Y],equal(plus(uminus(Y),plus(Y,X)),X))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(UMINUS(Y),PLUS(cY,X)),cX);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X,Y],equal(plus(plus(X,uminus(Y)),Y),X))). ***********************/
  /*** formerly: formula(forall([X,Y],equal(plus(minus(X,Y),Y),X))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(PLUS(X,UMINUS(Y)),cY),cX);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW formula(forall([X,Y],equal(plus(plus(X,Y),uminus(Y)),X))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(PLUS(X,Y),UMINUS(cY)),cX);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW formula(forall([X],equal(plus(X,uminus(X)),0))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(X,UMINUS(cX)),ZERO);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW formula(forall([X],equal(plus(uminus(X),X),0))). ***********************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(UMINUS(X),cX),ZERO);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(plus(X,0),X))). **********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(X,ZERO),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(plus(0,X),X))). **********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(PLUS(ZERO,X),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW3 formula(forall([X],lesseq(X,X))). ***********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = LESSEQ(X,cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],less(X,plus(X,1)))). ***********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  /* cf. "foldfg.h":93,179, "foldfg.c":81,121, and "dfgparser.c":1949:
     declare "1" additionally as function symbol with arity 0 */
  NewAtom = LESS(X,PLUS(cX,ONE));
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW formula(forall([X],less(plus(X,-1),X))). ***********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  /* cf. "foldfg.h":93,179, "foldfg.c":81,121, and "dfgparser.c":1949:
     declare "1" additionally as function symbol with arity 0 */
  NewAtom = LESS(PLUS(X,MINUSONE),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X,Y],or(less(X,Y),less(Y,X),equal(X,Y)))). **/
  /*** formerly: formula(forall([X,Y],implies(and(lesseq(X,Y),lesseq(Y,X)),equal(X,Y)))). **/
  /* <=> formula(forall([X,Y],or(not(X<=Y),not(Y<=X),X=Y)))*/
  NewClause = clause_CreateBody(3);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = LESS(X,Y);
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  
  NewAtom = LESS(cY,cX);
  clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

  NewAtom = EQUAL(cX,cY);
  clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

  clause_SetNumOfAnteLits(NewClause, 0);
  clause_SetNumOfSuccLits(NewClause, 3); 

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW4 formula(forall([X,Y],or(lesseq(X,Y),less(plus(Y,Z),plus(X,Z))))). **/
  NewClause = clause_CreateBody(2);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();
  Z = term_CreateStandardVariable();

  NewAtom = LESSEQ(X,Y);
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  
  NewAtom = LESS(PLUS(cY,Z),PLUS(cX,cZ));
  clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

  clause_SetNumOfAnteLits(NewClause, 0);
  clause_SetNumOfSuccLits(NewClause, 2); 

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW4 formula(forall([X,Y],or(lesseq(X,Y),less(plus(Z,Y),plus(Z,X))))). **/
  NewClause = clause_CreateBody(2);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();
  Z = term_CreateStandardVariable();

  NewAtom = LESSEQ(X,Y);
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  
  NewAtom = LESS(PLUS(Z,cY),PLUS(cZ,cX));
  clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

  clause_SetNumOfAnteLits(NewClause, 0);
  clause_SetNumOfSuccLits(NewClause, 2); 

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW4 formula(forall([X,Y],or(lesseq(X,Y),less(uminus(X),uminus(Y))))). **/
  NewClause = clause_CreateBody(2);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = LESSEQ(X,Y);
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  
  NewAtom = LESS(UMINUS(cX),UMINUS(cY));
  clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

  clause_SetNumOfAnteLits(NewClause, 0);
  clause_SetNumOfSuccLits(NewClause, 2); 

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formerly: formula(forall([X],equal(uminus(X),minus(0,X)))). *************************/
  /*** leads to non-termination with the new defn of precious clauses!
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  
  NewAtom = EQUAL(UMINUS(X),MINUS(ZERO,cX));
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  ***/
  /*******************************************************************************/


  /*** formula(forall([X],equal(times(X,0),0))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(X,ZERO),ZERO);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(times(0,X),0))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(ZERO,X),ZERO);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(times(X,1),X))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(X,ONE),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(times(1,X),X))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(ONE,X),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW5 formula(forall([X],equal(times(X,-1),uminus(X)))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(X,MINUSONE),UMINUS(cX));
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** NEW5 formula(forall([X],equal(times(-1,X),uminus(X)))). *********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(TIMES(MINUSONE,X),UMINUS(cX));
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(divide_nat(X,1),X))). ********************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();
  
  NewAtom = EQUAL(DIVIDENAT(X,ONE),cX);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X],equal(remainder(X,1),0))). *****************************/
  NewClause = clause_CreateBody(1);
  X = term_CreateStandardVariable();

  NewAtom = EQUAL(REMAINDER(X,ONE),ZERO);
  
  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(forall([X,Y],implies(evaleq(X,Y),equal(X,Y)))). **/
  NewClause = clause_CreateBody(2);
  X = term_CreateStandardVariable();
  Y = term_CreateStandardVariable();

  NewAtom = EVALEQ(X,Y);
  clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));
  
  NewAtom = EQUAL(cX,cY);
  clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

  clause_SetNumOfAnteLits(NewClause, 1);
  clause_SetNumOfSuccLits(NewClause, 1); 

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(evaleq(0,0)). *********************************/
  NewClause = clause_CreateBody(1);

  NewAtom = EVALEQ(ZERO,ZERO);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  /*** formula(evaleq(1,1)). *********************************/
  NewClause = clause_CreateBody(1);

  NewAtom = EVALEQ(ONE,ONE);

  clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
  clause_SetNumOfSuccLits(NewClause, 1);

  StandardAxioms = list_Cons(NewClause, StandardAxioms);
  /*******************************************************************************/


  if (TheoryFlag == flag_DPCTINT) {

    /*** formula(forall([X],isint(X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(X);

    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    StandardAxioms = list_Cons(NewClause, StandardAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],israt(X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = ISRAT(X);

    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    StandardAxioms = list_Cons(NewClause, StandardAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],equal(floor(X),X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = EQUAL(FLOOR(X),cX);
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    StandardAxioms = list_Cons(NewClause, StandardAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],equal(ceiling(X),X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = EQUAL(CEILING(X),cX);
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    StandardAxioms = list_Cons(NewClause, StandardAxioms);
    /*******************************************************************************/

  } else if ((TheoryFlag == flag_DPCTRAT || TheoryFlag == flag_DPCTREAL)
             && clause_ClauseListContainsSubsortOps(*ProblemClauses)) {

    if (TheoryFlag == flag_DPCTRAT) {

      /*** formula(forall([X],israt(X))). *********************************/
      NewClause = clause_CreateBody(1);
      X = term_CreateStandardVariable();

      NewAtom = ISRAT(X);

      clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
      clause_SetNumOfSuccLits(NewClause, 1);

      StandardAxioms = list_Cons(NewClause, StandardAxioms);
      /*******************************************************************************/

    } else {

      /*** formula(forall([X],implies(isint(X),israt(X)))). *********************************/
      NewClause = clause_CreateBody(2);
      X = term_CreateStandardVariable();

      NewAtom = ISINT(X);
      clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(cX);
      clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

      clause_SetNumOfAnteLits(NewClause, 1);
      clause_SetNumOfSuccLits(NewClause, 1);

      ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
      /*******************************************************************************/

      /*** formula(forall([X],implies(israt(X),israt(uminus(X))))). *********************************/
      NewClause = clause_CreateBody(2);
      X = term_CreateStandardVariable();

      NewAtom = ISRAT(X);
      clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(UMINUS(cX));
      clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

      clause_SetNumOfAnteLits(NewClause, 1);
      clause_SetNumOfSuccLits(NewClause, 1);

      ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
      /*******************************************************************************/


      /*** formula(forall([X],implies(and(israt(X),israt(Y)),israt(plus(X,Y))))). *********************************/
      NewClause = clause_CreateBody(3);
      X = term_CreateStandardVariable();
      Y = term_CreateStandardVariable();

      NewAtom = ISRAT(X);
      clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(Y);
      clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(PLUS(cX,cY));
      clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

      clause_SetNumOfAnteLits(NewClause, 2);
      clause_SetNumOfSuccLits(NewClause, 1);

      ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
      /*******************************************************************************/


      /*** formula(forall([X],implies(and(israt(X),israt(Y)),israt(times(X,Y))))). *********************************/
      NewClause = clause_CreateBody(3);
      X = term_CreateStandardVariable();
      Y = term_CreateStandardVariable();

      NewAtom = ISRAT(X);
      clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(Y);
      clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

      NewAtom = ISRAT(TIMES(cX,cY));
      clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

      clause_SetNumOfAnteLits(NewClause, 2);
      clause_SetNumOfSuccLits(NewClause, 1);

      ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
      /*******************************************************************************/

    }


    /*** formula(forall([X],isint(floor(X)))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(FLOOR(X));
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],isint(ceiling(X)))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(CEILING(X));
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],lesseq(floor(X),X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESSEQ(FLOOR(X),cX);
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],lesseq(X,ceiling(X)))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESSEQ(X,CEILING(cX));
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X],less(plus(X,-1),floor(X)))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESS(PLUS(X,MINUSONE),FLOOR(cX));
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],less(X,floor(plus(X,1))))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESS(X,FLOOR(PLUS(cX,ONE)));
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],less(ceiling(X),plus(X,1)))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESS(CEILING(X),PLUS(cX,ONE));
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],less(ceiling(plus(X,-1)),X))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = LESS(CEILING(PLUS(X,MINUSONE)),cX);
  
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(floor(X),X)))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(FLOOR(cX),cX);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(ceiling(X),X)))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(CEILING(cX),cX);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(plus(X,floor(Y)),floor(plus(X,Y)))))). *********************************/
    /*
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(PLUS(cX,FLOOR(Y)),FLOOR(PLUS(cX,cY)));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    */
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(plus(floor(Y),X),floor(plus(Y,X)))))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(PLUS(FLOOR(Y),cX),FLOOR(PLUS(cY,cX)));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(plus(X,ceiling(Y)),ceiling(plus(X,Y)))))). *********************************/
    /*
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(PLUS(cX,CEILING(Y)),CEILING(PLUS(cX,cY)));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    */
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),equal(plus(ceiling(Y),X),ceiling(plus(Y,X)))))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = EQUAL(PLUS(CEILING(Y),cX),CEILING(PLUS(cY,cX)));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(isint(X),isint(uminus(X))))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(UMINUS(cX));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 1);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(and(isint(X),isint(Y)),isint(plus(X,Y))))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(Y);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(PLUS(cX,cY));
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 2);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X,Y],implies(and(isint(plus(X,Y)),isint(uminus(Y))),isint(X)))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(PLUS(X,Y));
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(UMINUS(cY));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(cX);
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 2);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X,Y],implies(and(isint(plus(X,Y)),isint(uminus(X))),isint(Y)))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(PLUS(X,Y));
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(UMINUS(cX));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(cY);
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 2);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/


    /*** formula(forall([X],implies(and(isint(X),isint(Y)),isint(times(X,Y))))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = ISINT(X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(Y);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreateNegative(NewAtom, NewClause));

    NewAtom = ISINT(TIMES(cX,cY));
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 2);
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X,Y],or(less(Y,X),lesseq(floor(X),floor(Y))))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = LESS(Y,X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESSEQ(FLOOR(cX),FLOOR(cY));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 0);
    clause_SetNumOfSuccLits(NewClause, 2);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X,Y,Z],or(less(Z,Y),less(Y,X),lesseq(floor(X),floor(Z))))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();
    Z = term_CreateStandardVariable();

    NewAtom = LESS(Z,Y);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESS(cY,X);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESSEQ(FLOOR(cX),FLOOR(cZ));
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 0);
    clause_SetNumOfSuccLits(NewClause, 3);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X,Y],or(less(Y,X),lesseq(ceiling(X),ceiling(Y))))). *********************************/
    NewClause = clause_CreateBody(2);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();

    NewAtom = LESS(Y,X);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESSEQ(CEILING(cX),CEILING(cY));
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 0);
    clause_SetNumOfSuccLits(NewClause, 2);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X,Y,Z],or(less(Z,Y),less(Y,X),lesseq(ceiling(X),ceiling(Z))))). *********************************/
    NewClause = clause_CreateBody(3);
    X = term_CreateStandardVariable();
    Y = term_CreateStandardVariable();
    Z = term_CreateStandardVariable();

    NewAtom = LESS(Z,Y);
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESS(cY,X);
    clause_SetLiteral(NewClause, 1, clause_LiteralCreate(NewAtom, NewClause));

    NewAtom = LESSEQ(CEILING(cX),CEILING(cZ));
    clause_SetLiteral(NewClause, 2, clause_LiteralCreate(NewAtom, NewClause));

    clause_SetNumOfAnteLits(NewClause, 0);
    clause_SetNumOfSuccLits(NewClause, 3);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X],equal(floor(uminus(X)),uminus(ceiling(X))))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = EQUAL(FLOOR(UMINUS(X)),UMINUS(CEILING(cX)));
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

    /*** formula(forall([X],equal(ceiling(uminus(X)),uminus(floor(X))))). *********************************/
    NewClause = clause_CreateBody(1);
    X = term_CreateStandardVariable();

    NewAtom = EQUAL(CEILING(UMINUS(X)),UMINUS(FLOOR(cX)));
    
    clause_SetLiteral(NewClause, 0, clause_LiteralCreate(NewAtom, NewClause));
    clause_SetNumOfSuccLits(NewClause, 1);

    ExtensionAxioms = list_Cons(NewClause, ExtensionAxioms);
    /*******************************************************************************/

  }


  /* establish common properties for all newly generated clauses */
  for (Scan = StandardAxioms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    clause_InitSplitData((CLAUSE) list_Car(Scan));
    clause_OrientAndReInit((CLAUSE) list_Car(Scan), prfs_Store(Search), prfs_Precedence(Search));
    clause_SetTheoryStandardAxiom((CLAUSE) list_Car(Scan));
    clause_RemoveFlag((CLAUSE) list_Car(Scan), CONCLAUSE);

#ifdef CHECK
    clause_Check((CLAUSE) list_Car(Scan), prfs_Store(Search), prfs_Precedence(Search));
#endif
  }

  /* add them to the problem clauses */
  for (Scan = StandardAxioms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    *ProblemClauses = list_Cons(list_Car(Scan), *ProblemClauses);
  }

  /* delete auxilliary list without elements itself */
  list_Delete(StandardAxioms);

  /* establish common properties for all newly generated clauses */
  for (Scan = ExtensionAxioms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    clause_InitSplitData((CLAUSE) list_Car(Scan));
    clause_OrientAndReInit((CLAUSE) list_Car(Scan), prfs_Store(Search), prfs_Precedence(Search));
    clause_SetTheoryExtensionAxiom((CLAUSE) list_Car(Scan));
    clause_RemoveFlag((CLAUSE) list_Car(Scan), CONCLAUSE);

#ifdef CHECK
    clause_Check((CLAUSE) list_Car(Scan), prfs_Store(Search), prfs_Precedence(Search));
#endif
  }

  /* add them to the problem clauses */
  for (Scan = ExtensionAxioms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    *ProblemClauses = list_Cons(list_Car(Scan), *ProblemClauses);
  }

  /* delete auxilliary list without elements itself */
  list_Delete(ExtensionAxioms);
}
/* < SZ */

static LIST top_GetLiteralsForSplitting(CLAUSE Clause, BOOL SuccedentOnly)
/**************************************************************
  INPUT:   A clause and a flag whether only succedent literals should be
           considered
  RETURNS: A list of literal indices where every single
           literal doesn't share any variables with other literals.
           So these literals can be split as units.
           If SuccedentOnly is TRUE, only succedent lits are considered
***************************************************************/
{
  LIST* Variables;  /* An array, mapping literal index to list of variables */
  int   i, j, EndIndex;
  BOOL  Stop;
  LIST  Failed, Result;

  Result   = list_Nil();
  EndIndex = (SuccedentOnly ? clause_FirstSuccedentLitIndex(Clause) : clause_FirstLitIndex());

  /* Special case: clause is ground, so return all literals */
  if (clause_IsGround(Clause)) {
    for (i = clause_LastLitIndex(Clause); i >= EndIndex; i--)
      /* hier ist der fehler */
      Result = list_Cons((POINTER)i, Result);
    return Result;
  }

  Variables = memory_Malloc(sizeof(LIST) * clause_Length(Clause));
  /* Initialize the array */
  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(Clause); i++)
    Variables[i] = term_VariableSymbols(clause_GetLiteralAtom(Clause, i));

  /* <Failed> is the set of literals that share variables with other literals */
  Failed = list_Nil();
  for (i = clause_LastLitIndex(Clause); i >= EndIndex; i--) {
    if (list_Empty(Variables[i]))
      Result = list_Cons((POINTER)i, Result);
    else if (!list_PointerMember(Failed, (POINTER)i)) {
      /* We don't know yet whether the literal shares variables */
      Stop = FALSE;
      for (j = clause_FirstLitIndex();
	   j <= clause_LastLitIndex(Clause) && !Stop; j++) {
	if (i != j && list_HasIntersection(Variables[i], Variables[j])) {
	  Stop   = TRUE;  /* Literal isnt candidate for "optimal" splitting */
	  Failed = list_Cons((POINTER)i, Failed);
	  Failed = list_Cons((POINTER)j, Failed);
	}
      }
      if (!Stop)
	Result = list_Cons((POINTER)i, Result);
    }
  }
  
  /* Cleanup */
  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(Clause); i++)
    list_Delete(Variables[i]);
  memory_Free(Variables, sizeof(LIST) * clause_Length(Clause));
  list_Delete(Failed);
  return Result;
}

/* SZ > */
static int top_GetOptimalTheorySplitLiteralIndex(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proofsearch object and a clause.
  RETURNS: The index of the first ground literal from <Clause>.
	   The literal mustn't share any variables with other literals.
	   If the clause doesn't have a suitable literal, a negative
	   number is returned.
***************************************************************/
{
  LIST    SplitLits, Scan;
  LITERAL Literal;
  int     Result;
  
  /* Consider also negative literals */
  SplitLits    = top_GetLiteralsForSplitting(Clause, FALSE);
  Result       = -1;
  
  /* search the first theory literal (i.e. with theory head) */
  for (Scan=SplitLits; (!list_Empty(Scan) && (Result == -1)); Scan = list_Cdr(Scan)) {
    Literal = clause_GetLiteral(Clause, (int)list_Car(Scan));
    if (!clause_LiteralContainsVariables(Literal))
      Result = (int)list_Car(Scan);
  }

  list_Delete(SplitLits);
  return Result;
}
/* < SZ */

static int top_GetOptimalSPASSSplitLiteralIndex(PROOFSEARCH Search, CLAUSE Clause,
						BOOL Usables)
/**************************************************************
  INPUT:   A proofsearch object, a clause and a boolean flag.
  RETURNS: The index of the positive literal from <Clause>
           with the greatest number of instances (maybe 0) within
	   the WorkedOff/Usable sets of the proof search object.
	   The literal mustn't share any variables with other literals.
	   If the clause doesn't have a suitable literal, a negative
	   number is returned.
  EFFECT:  If <Usables> is FALSE, only the number of instances
	   within the WorkedOff set is considered, otherwise
	   the number of instances within both clause sets is considered.
***************************************************************/
{
  LIST    SplitLits;
  LITERAL Literal;
  NAT     Count, MaxInstances;
  int     Result;

  MaxInstances = -1;
  SplitLits    = top_GetLiteralsForSplitting(Clause, 
		     flag_GetFlagValue(prfs_Store(Search), flag_SPLITHEURISTIC) == flag_SPLITHEURISTICSTANDARD);
  Result       = -1;

  for ( ; !list_Empty(SplitLits); SplitLits = list_Pop(SplitLits)) {
    Literal = clause_GetLiteral(Clause, (int)list_Car(SplitLits));
    /* Count the number of instances */
    Count = prfs_GetNumberOfInstances(Search, Literal, Usables);
    if (Count > MaxInstances) {
      Result = (int)list_Car(SplitLits);
      MaxInstances = Count;
    }
  }
  return Result;
}

/* SZ > */
static int top_GetOptimalSplitLiteralIndex(PROOFSEARCH Search, CLAUSE Clause,
					   BOOL Usables, int *SplitHeuristicCounter)
/****************************************************************************
  INPUT:   A proofsearch object, a clause and a boolean flag, the 
           split heuristic counter.
  RETURNS: The result of either <top_GetOptimalSPASSSplitLiteralIndex> 
           or <top_GetOptimalTheoryplitLiteralIndex>.
  EFFECT:  Depending on the current value of <SplitHeuristicCounter> and
           the flag <DpcSTRatio> either the SPASS or the theory split
           heuristic is invoked.
****************************************************************************/
{
  int LitIndex;
  int heuristic;
  
  /* do not split auxilliary clauses (needed for theory instantiation */
  if (!clause_IsThGroundClause(Clause))
    LitIndex = -1;
  else if (clause_ResultedFromIntOrdExp(Clause))
    LitIndex =-1;
  /* do not split unit clauses */
  else if (clause_Length(Clause) < 2)
    LitIndex = -1;
  else {
    if (flag_GetFlagValue(prfs_Store(Search), flag_DPCSTRATIO) == 0)
      /* always apply SPASS heuristic */
      heuristic = 0;
    else if (((*SplitHeuristicCounter) % flag_GetFlagValue(prfs_Store(Search),
							   flag_DPCSTRATIO)) == 0)
      /* it's theory heuristic's turn */
      heuristic = 1;
    else
      /* it's SPASS heuristic's turn */
      heuristic = 0;
  
    /* apply chosen heuristic */
    if (heuristic == 0)
      LitIndex = top_GetOptimalSPASSSplitLiteralIndex(Search, Clause, Usables);
    else
      LitIndex = top_GetOptimalTheorySplitLiteralIndex(Search, Clause);

    if (LitIndex >= 0) {
      /* split was successful */
      ++(*SplitHeuristicCounter);
    }
  }

  return LitIndex;
}
/* < SZ */

/* SZ > */
int top_GetTheorySplitLiterals(CLAUSE Clause, LIST *SplitLiterals)
/****************************************************************************
  INPUT:   A clause and a list used as return value.
  RETURNS: 1 if there are ground literals that could be used for splitting, 
           -1 otherwise.
  EFFECT:  Selects all ground literals from the clause that aim to be 
           good to be split for the theory split heurstic.
****************************************************************************/
{
  LIST SplitLits, Scan2;
  LITERAL Literal;
  int Length;

  *SplitLiterals = list_Nil();
  SplitLits = top_GetLiteralsForSplitting(Clause, FALSE);

  Length = 0;
  for(Scan2=SplitLits; !list_Empty(Scan2); Scan2=list_Cdr(Scan2)) {
    Literal = clause_GetLiteral(Clause, (int)(list_Car(Scan2)));
    if (!clause_LiteralContainsVariables(Literal)) {
      *SplitLiterals = list_Cons(Literal, *SplitLiterals);
      ++Length;
    }
  }
  list_Delete(SplitLits);

  if ((Length > 0) && (Length == clause_Length(Clause))) {
    /* clause is purely ground (this code snippet is needed if the function
       is inokved from <top_LazyReductionSelectGivenComputeDerivables>) */
    list_Delete(*SplitLiterals);
    *SplitLiterals = NULL;
    
    return -1;
  }
  
  if (list_Empty(*SplitLiterals))
    return -1;
  else
    return 1;
}
/* < SZ */

/* SZ > */
static CLAUSE top_GetPowerfulTheorySplitClause(PROOFSEARCH Search, LIST *SplitLiterals)
/**********************************************************************************
  INPUT:   A proofsearch object and a pointer to a literal
           list which is used as return value.
  RETURNS: A clause from the usable set that was selected as given clause.
           Iff no suitable clause was found NULL is returned and <*SplitLiterals>
	   is set to NULL.
	   Iff a suitable clause was found, this clause is returned and
	   <*LitIndex> is set to the list of literals, that are split off
           and form the left part of the split clause
***********************************************************************************/
{
  LIST    Scan, SplitLits, TheoryLits, Scan2;
  int     MaxNrTheoryLits, NrTheoryLits, BestFirstTheoryLit;
  CLAUSE  Clause, OptimalClause;
  LITERAL Literal;
  
  OptimalClause      = NULL;
  *SplitLiterals     = NULL;
  MaxNrTheoryLits    = -1;
  BestFirstTheoryLit = -1;
  
#ifdef __DEBUG_SPLITTING__
  printf("Theory split scheme: %d\n", flag_GetFlagValue(prfs_Store(Search),
							flag_DPCTHEORYSPLITSCHEME));
#endif

  /* distinguish between theory split scheme */
  switch(flag_GetFlagValue(prfs_Store(Search), flag_DPCTHEORYSPLITSCHEME)) {
  case 1:
    /* choose a clause such that
         * the clause has the highest number of ground literals, at least one
	 * the clause has at least one non-theory literal
	    ==> a "mixed" clause
	 * the ground theory part (i.e. EUF+LIA) is split off 
       [ possible extensions:
             * the theory literals must not be necessarily ground
	     * the universal closure of it is exported
	     * the maximal set of theory literals which is variable disjoint with the
               non-theory-literals and the remaining theory literals is exported ] */

    /* iterate over all clauses and take the one with the highest number of 
       ground literals with a theory head (EUF+LIA) */
    for (Scan = prfs_UsableClauses(Search); !list_Empty(Scan);
	 Scan = list_Cdr(Scan)) {
      Clause = list_Car(Scan);

      /* TEMPORARY changes */
      if (clause_HasSolvedConstraint(Clause)
	  && (clause_Length(Clause) > 1)
	  && !clause_ResultedFromIntOrdExp(Clause)
	  && !clause_IsThGroundClause(Clause)
/* 	  && !clause_IsFromArithmeticSimplification(Clause) */
	  ) {
	
	/* Get a list of splittable literal indices, not necessarily succedent ones */
	
	SplitLits = top_GetLiteralsForSplitting(Clause, FALSE);
	
	TheoryLits = list_Nil();
	for(Scan2=SplitLits; !list_Empty(Scan2); Scan2=list_Cdr(Scan2)) {
	  Literal = clause_GetLiteral(Clause, (int)(list_Car(Scan2)));
	  if (!clause_LiteralContainsVariables(Literal)) {
	    TheoryLits = list_Cons(Literal, TheoryLits);
	  }
	}
	
	NrTheoryLits = (int)list_Length(TheoryLits);
	list_Delete(SplitLits);
	list_Delete(TheoryLits);

	
	if ((MaxNrTheoryLits < NrTheoryLits) 
	    && (NrTheoryLits > 0) 
	    && (NrTheoryLits < clause_Length(Clause))
	    ) {
	  MaxNrTheoryLits = NrTheoryLits;
	  OptimalClause = Clause;
	}
      } /* <Clause> was a candidate for splitting */
    } /* foreach clause \in Us */
    
    if ((OptimalClause != NULL) && (MaxNrTheoryLits > 0)) {
      if (top_GetTheorySplitLiterals(OptimalClause, SplitLiterals) != 1) {
	misc_StartErrorReport();
	misc_ErrorReport("\n Error in finding theory split literals.");
	misc_FinishErrorReport();
      }
    } else {
      OptimalClause = NULL;
      *SplitLiterals = NULL;
    }
    break;
  default:
    misc_StartErrorReport();
    misc_ErrorReport("\n Unknown theory split scheme chosen.");
    misc_FinishErrorReport();

    break;
  } /* switch */
  
  return OptimalClause;
}
/* < SZ */

/* EK: hier lassen oder nach search.c oder nach rules-split.c? */
static CLAUSE top_GetPowerfulSPASSSplitClause(PROOFSEARCH Search, BOOL Usables,
					      LIST *SplitLiterals)
/**************************************************************
  INPUT:   A proofsearch object, a boolean flag and a pointer to a literal list
           which is used as return value.
  RETURNS: A clause from the usable set that was selected as given clause.
           Iff no suitable clause was found NULL is returned and <*LitIndex>
	   is set to NULL.
	   Iff a suitable clause was found, this clause is returned and
	   <*SplitLiterals> contains the "optimal" literal, that
           is a literal that can be split off as unit with a high reduction
           potential
  EFFECT:  This function selects a clause from the "usable" set and
           a literal that are "optimal" for the application of the splitting
	   rule with respect to the following criteria:
           1) the literal must occur in the succedent of the non-horn clause,
	   2) the literal mustn't share any variables with other literals,
           3) the clause must have a solved constraint,
	   4) the atom must have the maximum number of instances
              a) within the set of "workedoff" clauses, iff <Usables>=FALSE
              b) within the set of "usable" and "workedoff" clauses,
                 iff "Usables"=TRUE
	   5) the atom must have at least one instance in another clause.
***************************************************************/
{
  LIST   Scan, SplitLits;
  NAT    MaxNrOfInstances, NrOfInstances;
  CLAUSE Clause, OptimalClause;
  TERM   Atom;
  SHARED_INDEX WOIndex, UsIndex;
  /* SZ > */
  int SplitLiteral;
  /* < SZ */

  OptimalClause    = NULL;
  *SplitLiterals   = NULL;
  MaxNrOfInstances = 0;
  WOIndex          = prfs_WorkedOffSharingIndex(Search);
  UsIndex          = prfs_UsableSharingIndex(Search);
  /* SZ > */
  SplitLiteral     = -1;
  /* < SZ */

  /* Prepare the term stamp */
  if (term_StampOverflow(sharing_StampID(WOIndex)))
    sharing_ResetAllTermStamps(WOIndex);
  if (Usables && term_StampOverflow(sharing_StampID(UsIndex)))
    sharing_ResetAllTermStamps(UsIndex);
  term_StartStamp();

  for (Scan = prfs_UsableClauses(Search); !list_Empty(Scan);
       Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);

    if (clause_HasSolvedConstraint(Clause) && 
	clause_Length(Clause) > 1 &&
	/* SZ > */
	!clause_IsThGroundClause(Clause) &&
	!clause_ResultedFromIntOrdExp(Clause) &&
	/* < SZ */
	(!clause_IsHornClause(Clause) || 
	 (flag_GetFlagValue(prfs_Store(Search), flag_SPLITHEURISTIC) == flag_SPLITHEURISTICGROUND &&
	  clause_GetFlag(Clause, CONCLAUSE) &&  /* Split non-Horn clauses or input conjecture clauses */
	  clause_Depth(Clause) == 0))) {
      /* Get a list of splittable literal indices */
      SplitLits = top_GetLiteralsForSplitting(Clause, 
		     flag_GetFlagValue(prfs_Store(Search), flag_SPLITHEURISTIC) == flag_SPLITHEURISTICSTANDARD);
      for ( ; !list_Empty(SplitLits); SplitLits = list_Pop(SplitLits)) {
	LITERAL Literal;

	Literal = clause_GetLiteral(Clause, (int)list_Car(SplitLits));
	Atom    = clause_LiteralAtom(Literal);
 	if (!term_AlreadyVisited(Atom)) {
	  /* Don't visit atom more than once */
 	  term_SetTermStamp(Atom);
	  /* Count the number of instances */
	  NrOfInstances = prfs_GetNumberOfInstances(Search, Literal, Usables);
	  if (NrOfInstances > MaxNrOfInstances || OptimalClause == NULL ||
	      (NrOfInstances == MaxNrOfInstances &&
	       /* Prefer shorter clauses for splitting! */
	       clause_Length(Clause) < clause_Length(OptimalClause))) {
	    OptimalClause    = Clause;
	    MaxNrOfInstances = NrOfInstances;
	    /* SZ > */
	    SplitLiteral = (int)list_Car(SplitLits);
	    /* < SZ */
	  }
 	}
      }
    }
  }
  term_StopStamp();

  /* The splittable literal must have at least one instance to be useful */
  /* reducing other clauses. If <Usables> is TRUE, the literal must even */
  /* have two instances, since the literal of the given clause is in the */
  /* usable index, too.                                                  */
  if (OptimalClause != (CLAUSE)NULL &&
      (MaxNrOfInstances == 0 || (Usables && MaxNrOfInstances == 1)) &&
      clause_Depth(OptimalClause) != 0)
  {
    SplitLiteral  = -1;
    OptimalClause = NULL;
  }
  
  if ((SplitLiteral != -1) && (OptimalClause != NULL))
    *SplitLiterals = list_List(clause_GetLiteral(OptimalClause, SplitLiteral));
  else
    *SplitLiterals = NULL;
  
  return OptimalClause;
}

/* SZ > */
static CLAUSE top_GetPowerfulSplitClause(PROOFSEARCH Search, BOOL Usables,
					 LIST *SplitLits, int *SplitHeuristicCounter,
					 int *TheoryHeuristicApplied)
/****************************************************************************
  INPUT:   A proofsearch object, a clause and a boolean flag, a list of split
           literals which is used as a return value, the split heuristic
           counter and a flag indicating which heuristic was applied.
  RETURNS: The result of either <top_GetPowerfulSPASSSplitClause> 
           or <top_GetPowerfulTheorySplitClause>.
  EFFECT:  Depending on the current value of <SplitHeuristicCounter> and
           the flag <DpcSTRatio> either the SPASS or the theory split
           heuristic is invoked to get clause to be split.
  NOTE:    In flag.h the flag <flag_DPCSTRATIO> is set to have range 0..0.
           Thus, always the SPASS split heuristic is used; in the 
           implementation phase there was no evidence that a specialized
           split heuristic pays off; however, there is space for future 
           optimizations
****************************************************************************/
{
  CLAUSE SplitClause;
  int heuristic;
  
  if (flag_GetFlagValue(prfs_Store(Search), flag_DPCSTRATIO) == 0)
    /* always apply SPASS heuristic */
    heuristic = 0;
  else if (((*SplitHeuristicCounter) % flag_GetFlagValue(prfs_Store(Search),
							 flag_DPCSTRATIO)) == 0)
    /* it's theory heuristic's turn */
    heuristic = 1;
  else
    /* it's SPASS heuristic's turn */
    heuristic = 0;
  
  /* apply chosen heuristic */
  if (heuristic == 0)
    SplitClause = top_GetPowerfulSPASSSplitClause(Search, Usables, SplitLits);
  else
    SplitClause = top_GetPowerfulTheorySplitClause(Search, SplitLits);

  if (TheoryHeuristicApplied != NULL)
    *TheoryHeuristicApplied = heuristic;

  if (SplitClause != NULL)
    ++(*SplitHeuristicCounter);
  
  return SplitClause;
}
/* < SZ */

static LIST top_FullReductionSelectGivenComputeDerivables( PROOFSEARCH Search,
							   CLAUSE *SplitClause,
							   int *Counter, 
							   /*SZ>*/ 
							   int *SplitHeuristicCounter
							   /*<SZ*/ )
/**************************************************************
  INPUT:   A proof search object, a pointer to a clause resulting from a
           previous splitting step, and a pointer to an integer counter.
  RETURNS: A list of derived clauses.
  EFFECT:  In this function a clause is selected from the set of
           "usable" clauses. After a clause was selected as "given clause",
	   inferences between the given clause and the "worked off" clauses
	   are made. The selection works as follows:
	   1) If <*SplitClause> is not NULL, the split clause
	      is selected as "given clause". <*SplitClause> should result
	      from splitting
	   2) If <*SplitClause> is NULL, we try to find a clause that is
	      "optimal" for splitting. This is done by selecting a literal
	      <L> in a clause, such that <L> is variable-disjoint from
	      the rest of the clause, and the atom of <L> has the maximum
	      number of instances within the set of "usable" and "workoff"
	      clauses.
	   3) If the previous steps failed, a clause is selected by weight
	      or by depth, depending on the parameters "WDRatio", "PrefVar"
	      and "PrefCon". Then splitting is tried on the selected clause.
	      If the clause is a non-horn clause, we try to find a positive
	      literal <L> and a set of negative literals <N>, such that <N>
	      and <L> are variable disjoint from the rest of the clause.
***************************************************************/
{
  CLAUSE     GivenClause, TerminatorClause;
  LIST       Derivables, SplitLits;
  FLAGSTORE  Flags;
  PRECEDENCE Precedence;
  int        LitIndex;

  GivenClause = NULL;
  Derivables  = list_Nil();
  Flags       = prfs_Store(Search);
  Precedence  = prfs_Precedence(Search);

  /* 1) If the last given clause was split or if backtracking was applied, */
  /* then choose the clause resulting from the splitting step as        */
  /* given clause.                                                      */
  /* ATTENTION: Since the <SplitClause> might have been reduced since   */
  /* the time the variable was set, we have to check whether            */
  /* <SplitClause> is still element of the set of usable clauses.       */
  if (*SplitClause != NULL &&
      list_PointerMember(prfs_UsableClauses(Search), *SplitClause)) {
    GivenClause = *SplitClause;
  }
  
  *SplitClause = NULL;
  
  if (GivenClause == NULL) {
    if (prfs_SplitCounter(Search) != 0)
      /* 2) Try to find an "optimal" splitting clause, that doesn't share */
      /*    variables with any other literal.                             */
      
      GivenClause = top_GetPowerfulSplitClause(Search, FALSE, &SplitLits,
					       /*SZ>*/SplitHeuristicCounter,
					       NULL/*<SZ*/);
    if (GivenClause != NULL) {
      
      /* Found "optimal" split clause, so apply splitting */
      /* > SZ < */
      /* SplitLits = list_List(clause_GetLiteral(GivenClause, LitIndices)); */
      
      *SplitClause = prfs_DoSplittingAndInformDPC(Search, GivenClause, SplitLits);

      list_Delete(SplitLits);
      
    } else {
      /* 3) Splitting wasn't applied, so use the other strategies */
      if ((*Counter) % flag_GetFlagValue(Flags, flag_WDRATIO) == 0)
	GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search), Flags);
      else {
	if (flag_GetFlagValue(Flags, flag_PREFCON) != flag_PREFCONUNCHANGED)
	  GivenClause = top_SelectMinimalConWeightClause(prfs_UsableClauses(Search),
							 Flags);
	else
	  GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search),
						      Flags);
      }
      (*Counter)++; /* EK: hier lassen, oder eine Klammerebene nach aussen? */
    }
  }

  if (*SplitClause == NULL && prfs_SplitCounter(Search) != 0) {
    /* Try to find the "optimal" literal for splitting the clause. */
    /* This makes sense for a clause that is the right part of a   */
    /* splitting step.                                             */
    LitIndex = top_GetOptimalSplitLiteralIndex(Search, GivenClause, FALSE,
					       /*SZ>*/SplitHeuristicCounter/*<SZ*/);
    
    if (LitIndex >= 0) {
      SplitLits = list_List(clause_GetLiteral(GivenClause, LitIndex));
      *SplitClause = prfs_DoSplittingAndInformDPC(Search, GivenClause, SplitLits);
      
#ifdef __DEBUG_SPLITTING__
      printf("Splitting induced von GetOptimalSplitLits: %s\n", (*SplitClause != NULL) ? "successfully" : "unsuccessfully");
#endif

      list_Delete(SplitLits);
    } else {
      /* Optimal splitting wasn't possible, so try the old-style splitting. */
      /* Here a split is done if a positive literal doesn't share variables */
      /* with another POSITIVE literal. */
      *SplitClause = prfs_PerformSplitting(Search, GivenClause);
#ifdef __DEBUG_SPLITTING__
      printf("Ran performSplitting %s\n", (*SplitClause != NULL) ? "successfully" : "not successfully");
#endif
      fflush(stdout);
    }
  }

  prfs_ExtractUsable(Search, GivenClause);

  if (flag_GetFlagValue(Flags, flag_PGIVEN)) {
    fputs("\n\tGiven clause: ", stdout);
    clause_Print(GivenClause);
    fflush(stdout);
  }
  
  if (*SplitClause != NULL)
    Derivables = list_List(*SplitClause);
  else {
    /* No splitting was applied */
    if (flag_GetFlagValue(Flags, flag_RTER) != flag_RTEROFF) {
      clock_StartCounter(clock_REDUCTION);
      TerminatorClause = red_Terminator(GivenClause,
					flag_GetFlagValue(Flags, flag_RTER),
					prfs_WorkedOffSharingIndex(Search),
					prfs_UsableSharingIndex(Search), Flags,
					Precedence);
      clock_StopAddPassedTime(clock_REDUCTION);
      
      if (TerminatorClause != NULL) {
	/* An empty clause was derived by the "terminator" rule */
	Derivables = list_List(TerminatorClause);
	prfs_InsertUsableClause(Search, GivenClause);
      }
    }
    if (list_Empty(Derivables)) {
      /* No splitting was applied, no empty clause was found by terminator */
      clause_SelectLiteral(GivenClause, Flags);
      /*clause_SetSpecialFlags(GivenClause,ana_SortDecreasing());*/
      prfs_InsertWorkedOffClause(Search, GivenClause);
      clock_StartCounter(clock_INFERENCE);
      Derivables = inf_DerivableClauses(Search, GivenClause);
      clock_StopAddPassedTime(clock_INFERENCE);
    }
  }

  prfs_IncDerivedClauses(Search, list_Length(Derivables));

  return Derivables;
}


static LIST top_LazyReductionSelectGivenComputeDerivables( PROOFSEARCH Search,
							   CLAUSE *SplitClause,
							   int *Counter, 
							   /*SZ>*/ 
							   int *SplitHeuristicCounter
							   /*<SZ*/
							   )
/**************************************************************
  INPUT:   A proof search object, a pointer to a clause resulting from a
           previous splitting step, and a pointer to an integer counter.
  RETURNS: A list of derived clauses.
  EFFECT:  In this function a clause is selected from the set of
           "usable" clauses. After a clause was selected as "given clause",
	   inferences between the given clause and the "worked off" clauses
	   are made. Take a look at the description of the function
	   top_FullReduction... for more details.
	   This function is more complicated than the other function,
	   since clauses in the set of usable clauses may be reducible.
	   Because of this fact, the selection of the given clause
	   has to be done in a loop. After picking a "candidate" clause
	   the clause is inter-reduced with the "worked off" set.
	   If the candidate still exists after the reduction it is
	   selected as given clause, else another usable clause is picked
	   as candidate.
***************************************************************/
{
  CLAUSE     GivenClause, TerminatorClause;
  LIST       Derivables;
  FLAGSTORE  Flags;
  PRECEDENCE Precedence;
  int        LitIndex;

  /* SZ > */
  LIST SplitLiteralsOfClause;
  int  TheoryHeuristicApplied;
  /* < SZ */

  GivenClause            = (CLAUSE)NULL;
  TerminatorClause       = (CLAUSE)NULL;
  Derivables             = list_Nil();
  Flags                  = prfs_Store(Search);
  Precedence             = prfs_Precedence(Search);
  TheoryHeuristicApplied = 0;

  while (GivenClause == (CLAUSE)NULL &&
	 !list_Empty(prfs_UsableClauses(Search))) {
    /* The selected clause may be redundant */

    if (*SplitClause != NULL &&
	list_PointerMember(prfs_UsableClauses(Search), *SplitClause))
      GivenClause = *SplitClause;

    *SplitClause  = NULL;

    /* Try selecting a clause that is optimal for splitting */
    if (GivenClause == NULL) {
      if (prfs_SplitCounter(Search) != 0) {
	GivenClause = top_GetPowerfulSplitClause(Search, FALSE, &SplitLiteralsOfClause,
						 /*SZ>*/SplitHeuristicCounter,
						 &TheoryHeuristicApplied/*<SZ*/);
	/* The value of <SplitLiteralsOfClause> isn't used here. */
	if (SplitLiteralsOfClause != NULL)
	  list_Delete(SplitLiteralsOfClause);
      }

      if (GivenClause == NULL) {
	/* No optimal clause for splitting found */
	if ((*Counter) % flag_GetFlagValue(Flags, flag_WDRATIO) == 0)
	  GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search), Flags);
	else {
	  if (flag_GetFlagValue(Flags, flag_PREFCON) != flag_PREFCONUNCHANGED)
	    GivenClause = top_SelectMinimalConWeightClause(prfs_UsableClauses(Search),
							   Flags);
	  else
	    GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search),
							Flags);
	}
	(*Counter)++;
      }
    }
    prfs_ExtractUsable(Search, GivenClause);

    /* Reduce the selected clause */
    clock_StartCounter(clock_REDUCTION);
    GivenClause = red_CompleteReductionOnDerivedClause(Search, GivenClause, 
						       red_WORKEDOFF);
    clock_StopAddPassedTime(clock_REDUCTION);
  }

  if (GivenClause == (CLAUSE)NULL)
    /* Set of usable clauses is empty */
    return list_Nil();

  
  if (clause_IsEmptyClause(GivenClause)) {
    Derivables = list_List(GivenClause);
    return Derivables;
  }
  else {
    /* Reduce Workedoff clauses with selected clause */
    clock_StartCounter(clock_REDUCTION);
    Derivables = red_BackReduction(Search, GivenClause, red_WORKEDOFF);
    clock_StopAddPassedTime(clock_REDUCTION);
  }

  /* Print selected clause */
  if (flag_GetFlagValue(Flags, flag_PGIVEN)) {
    fputs("\n\tGiven clause: ", stdout);
    clause_Print(GivenClause); 
    fflush(stdout);
  }     

  /* Try splitting */
  if (prfs_SplitCounter(Search) != 0) {
    /* > SZ < */

    /* First try "optimal" splitting on the selected clause */
    LIST SplitLits;

    if (!TheoryHeuristicApplied)
      LitIndex = top_GetOptimalSPASSSplitLiteralIndex(Search, GivenClause, FALSE);
    else
      LitIndex = top_GetTheorySplitLiterals(GivenClause, &SplitLits);

    if (LitIndex >= 0) {
      if (!TheoryHeuristicApplied)
	SplitLits = list_List(clause_GetLiteral(GivenClause, LitIndex));

      *SplitClause = prfs_DoSplittingAndInformDPC(Search, GivenClause, SplitLits);
      list_Delete(SplitLits);
    } else {
      /* Try the old splitting that allows negative literals   */
      /* sharing variables with the selected positive literal. */
      *SplitClause = prfs_PerformSplitting(Search, GivenClause);
    }
  }
  
  if (*SplitClause != NULL) {
    Derivables = list_Cons(*SplitClause, Derivables);
  } else {
    /* Try terminator reduction only for a clause that wasn't splitted. */
    if (flag_GetFlagValue(Flags, flag_RTER) != flag_RTEROFF) {
      TerminatorClause = red_Terminator(GivenClause,
					flag_GetFlagValue(Flags, flag_RTER),
					prfs_WorkedOffSharingIndex(Search),
					prfs_UsableSharingIndex(Search),
					Flags, Precedence);
      if (TerminatorClause != NULL) {
	Derivables = list_Cons(TerminatorClause, Derivables);
	prfs_InsertUsableClause(Search, GivenClause);
      }
    }
    if (TerminatorClause == (CLAUSE)NULL) {   
      clause_SelectLiteral(GivenClause, Flags);
      /* clause_SetSpecialFlags(GivenClause,ana_SortDecreasing());*/
      prfs_InsertWorkedOffClause(Search, GivenClause);
      clock_StartCounter(clock_INFERENCE);
      Derivables = list_Nconc(Derivables,
			      inf_DerivableClauses(Search, GivenClause));
      clock_StopAddPassedTime(clock_INFERENCE);
    }
  }
  
  prfs_IncDerivedClauses(Search, list_Length(Derivables));

  return Derivables;
}

/* SZ > */
static int top_letSMTContribute(PROOFSEARCH Search, FLAGSTORE Flags,
				SMT_RESULT *smt_result, LIST *EmptyClauses,
				CLAUSE *HighestLevelEmptyClause, int *smt_sat)
/*********************************************************************
  INPUT:   A proof search object, a flag store, an SMT result, a list of clauses, 
           a clause, and a flag whether the SMT procedure yield 
           the result "satisfiable".
  RETURNS: Whether the application was successful, 0, or not, 1.
  EFFECTS: The result <smt_result> is handled, i.e. has an effect to 
           the search process.
           For unsatisfiability  the following is done: an empty 
           clause is created and inserted into <EmptyClauses>. 
           Its set of parent clauses and split data are assigned 
           according to the information of the parent clauses.
           Then its dept is compared to that of <HighestEmptyClause>. 
           If the dept of the newly generated empty clause is 
           higher (splitlevel is smaller) then it will become the 
           new <HighestEmptyClause>. 
           For "Sat(p)" the flag <smt_sat> is set (if != NULL). This
           is used if an urgent check was performed (either due to
           SPASS saturation or due to an open branch): it affects the
           final result of SPASS+T (since we're dealing with the 
           logic of arithmetic and free function symbols).
**********************************************************************/
{
  CLAUSE EmptyClause;
  int result;
  POSITION *p_spass;

  /* mark the result as seen */
  smt_result_markSeen(smt_result);
  
  /* possibly set the sat flag */
  if (smt_sat != NULL)
    *smt_sat = 0; /* no satisfiability (so far) */
  
  /* get the current SPASS position */
  p_spass = prfs_getPositionOfCurrentLevelInCurrentConfiguration(Search);

  if (smt_result_isUnsat(smt_result)) {
    /****************************/
    /* unsatisfiability result  */
    /****************************/
    
#ifdef CHECK
    /* check whether explanation was required */

    if ((flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) == flag_DPCEVIDENCEOFF) &&
	(smt_result_isUnsatWithEvidence(smt_result))) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In top_letSMTContribute: ");
      misc_ErrorReport("Error: got result with explanation even though it was not expected");
      misc_FinishErrorReport();
    }

    if ((flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) == flag_DPCEVIDENCEON) &&
	(smt_result_isUnsatWithoutEvidence(smt_result))) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In top_letSMTContribute: ");
      misc_ErrorReport("Error: got result without explanation even though it was required");
      misc_FinishErrorReport();
    }
#endif

#ifdef SMT_DEBUG
    if (smt_result_isUnsatWithEvidence(smt_result)) {
      smt_debug_start();
      printf("Got unsatisfiability with evidence result.\n");
      smt_debug_end();
    } else {
      smt_debug_start();
      printf("Got unsatisfiability without evidence result.\n");
      smt_debug_end();
    }
#endif
#ifdef SMT_LOG
    if (smt_result_isUnsatWithEvidence(smt_result)) {
      dpconnect_logMessage(Search, 1, "Got unsatisfiability with evidence result.\n");
    } else {
      dpconnect_logMessage(Search, 1, "Got unsatisfiability without evidence result.\n");
    }
#endif
    
    /* check whether result is relevant */
    if (smt_result_isRelevant(smt_result, Search, p_spass)) {
      EmptyClause = smt_result_generateEmptyClause(Search, smt_result);

      /* insert clause to <*EmptyClause> */
      *EmptyClauses = list_Cons(EmptyClause, *EmptyClauses);

      if (flag_GetFlagValue(Flags, flag_PEMPTYCLAUSE)) {
	fputs("\nEmptyClause: ", stdout);
	clause_Print(EmptyClause);
      }

#ifdef SMT_DEBUG
      smt_debug_start();
      printf("\nInserted an empty clause due to theory inconsistency.\n");
      clause_Print(EmptyClause);
      printf("\n\n");
      smt_debug_end();      
#endif
#ifdef SMT_LOG
      dpconnect_logMessage(Search, 0, 
			   "\tInserted an empty clause.\n");
#endif
      
      /* determine new highest empty clause*/
      if (clause_Exists(*HighestLevelEmptyClause)) {
	if (clause_SplitLevel(EmptyClause) < clause_SplitLevel(*HighestLevelEmptyClause))
	  *HighestLevelEmptyClause = EmptyClause;
      } else
	*HighestLevelEmptyClause = EmptyClause;
      
      EmptyClause = NULL;
    } else {
#ifdef SMT_DEBUG
      smt_debug_start();
      printf("This result was not relevant..\n");
      smt_debug_end();
#endif
#ifdef SMT_LOG
      dpconnect_logMessage(Search, 0, "\tThis result was not relevant.\n");
#endif      
    } /* unsatisfiability relevant ? */

    result = 0;  /* application successful */
  } else if (smt_result_isSat(smt_result)) {
    /****************************/
    /* satisfiability result    */
    /****************************/
#ifdef SMT_DEBUG
    smt_debug_start();
    printf("Got CONTROLSAT result.\n");
    if (flag_GetFlagValue(prfs_Store(Search),
			  flag_DPCNONGROUNDEXPORT) == flag_OFF)
      printf("\tNote however: The result is ignored: non-ground export disabled.\n");
    smt_debug_end();
#endif
#ifdef SMT_LOG
    dpconnect_logMessage(Search, 1, "Got CONTROLSAT result\n");
    if (flag_GetFlagValue(prfs_Store(Search),
			  flag_DPCNONGROUNDEXPORT) == flag_OFF)
      dpconnect_logMessage(Search, 0, "\tNote however: The result is ignored: non-ground export disabled.\n");
#endif
    /* this flag is later reached to the main function (via function "top_ProofSearch") */
    /* is only sound if SMT procedure has seen all input formulae which is only guaranteed
       if also non-ground formulae are exported; otherwise the result is ignored */
    if (flag_GetFlagValue(prfs_Store(Search),
			  flag_DPCNONGROUNDEXPORT) != flag_OFF)
      if (smt_result_isRelevant(smt_result, Search, p_spass))
	if (smt_sat != NULL)
	  *smt_sat = 1;
    
    result = 0;  /* application successful */
  } else if (smt_result_isUnknown(smt_result)) {
    /****************************/
    /* unknown result           */
    /****************************/

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("Got CONTROLUNKNOWN result.\n");
    smt_debug_end();
#endif
#ifdef SMT_LOG
    dpconnect_logMessage(Search, 1, 
			 "Got CONTROLUNKNOWN result.\n");
#endif
    result = 0;  /* application successful */
  } else {
    /* could not understand result */
    result = 1;  /* application not successful */
  }
  
  position_destroyPosition(p_spass);

  return result;
}
/* < SZ */

static PROOFSEARCH top_ProofSearch(PROOFSEARCH Search, LIST ProblemClauses,
				   FLAGSTORE InputFlags, LIST UserPrecedence, 
				   int *BoundApplied,
				   int *smt_sat)
/**************************************************************
  INPUT:   A proof search object, a list of clauses, a flag store
           containing the flags from the command line and from
           the input file, a list containing the precedence as
	   specified by the user, a pointer to an integer, and
           another pointer to an integer stating whether an 
           urgent invocation of the SMT procedure yield the
           result "satisfiable".
  RETURNS: The same proof search object
  EFFECTS: 
***************************************************************/
{
  LIST       Scan,EmptyClauses,Derivables;
  LIST       UsedEmptyClauses;
  CLAUSE     SplitClause,HighestLevelEmptyClause;
  FLAGSTORE  Flags;
  PRECEDENCE Precedence;
  int        Counter, ActBound, BoundMode, BoundLoops;
  /* SZ > */
  SMT_RESULT *smt_result;
  POSITION   *p_spass;
  CLAUSE     EmptyClause;
  int        SplitHeuristicCounter;
  /* < SZ */


  HighestLevelEmptyClause = (CLAUSE)NULL;
  UsedEmptyClauses        = list_Nil();
  EmptyClauses            = list_Nil();
  Derivables              = list_Nil();
  Flags                   = prfs_Store(Search);
  Precedence              = prfs_Precedence(Search);
  Counter                 = 1;
  /* SZ > */  
  p_spass                 = (POSITION*) NULL;
  EmptyClause             = (CLAUSE) NULL;
  SplitHeuristicCounter   = 1;
  *smt_sat                = 0;
  /* < SZ */

  clock_InitCounter(clock_REDUCTION);
  clock_InitCounter(clock_BACKTRACK);
  clock_InitCounter(clock_INFERENCE);
  
  /* (UW) */
  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
    
    clock_StartCounter(clock_SMT);
    /* give input clauses to SMT procedure */
    for (Scan=ProblemClauses; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      /* > SZ < */
      dpconnect_exportClause(Search, list_Car(Scan));
    }

    /* SZ > */
    /* tell SMT procedure that it got all input clauses; this 
       invokes the first call of the SMT procedure */
    dpconnect_gotAllProblemInputClauses(Search);
    /* < SZ */

    clock_StopAddPassedTime(clock_SMT);
  }
  
  /* Important ! Recomputes Weight ! */
  ana_AnalyzeProblem(Search, ProblemClauses);

  /* (UW) */
  /* VERY BAD TEMPORARY HACK! We reset a global variable of analyze.c
     back to false that seems to be incompatible with arithmetic
     simplification, but may have been set to true in ana_AnalyzeProblem.
     If this turns out to work, the interface of ana_AnalyzeProblem
     should be extended instead. */

  if (flag_GetFlagValue(Flags,flag_DPCTHEORY) >= flag_DPCTINT &&
      flag_GetFlagValue(Flags, flag_RARITH)) {
    ana_SORTRES = FALSE;
    ana_USORTRES = FALSE;
  }
  
  if (flag_GetFlagValue(Flags, flag_AUTO)) {
    prfs_InstallFiniteMonadicPredicates(Search, ProblemClauses, ana_FinMonPredList());
    ana_AutoConfiguration(ProblemClauses, Flags, Precedence);
    /* File and input flags have higher precedence */
    flag_TransferSetFlags(InputFlags, Flags);
  }

  /* SZ > */
#ifdef SMT_LOG
  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
    dpconnect_logSettings(Search);
  }
#endif
  /* < SZ */

  /* Rearrange automatically determined precedence according to user's specification. */
  symbol_RearrangePrecedence(Precedence, UserPrecedence);

  for (Scan = ProblemClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    clause_OrientAndReInit(list_Car(Scan), Flags, Precedence);

  /* Must be called after ana_AnalyzeProblem and Reorientation */
  ana_AnalyzeSortStructure(ProblemClauses, Flags, Precedence);

  if (flag_GetFlagValue(Flags, flag_AUTO)) {
    ana_ExploitSortAnalysis(Flags);
    /* File and input flags have higher precedence */
    flag_TransferSetFlags(InputFlags, Flags);
  }
  prfs_SetSplitCounter(Search, flag_GetFlagValue(Flags, flag_SPLITS));

  ActBound       = flag_GetFlagValue(Flags, flag_BOUNDSTART);
  BoundMode      = flag_GetFlagValue(Flags, flag_BOUNDMODE);
  BoundLoops     = flag_GetFlagValue(Flags, flag_BOUNDLOOPS);
  *BoundApplied  = -1;

  if (flag_GetFlagValue(Flags, flag_PPROBLEM)) {
    puts("");
#ifdef CHECK
    puts("--------------SPASS-START (with runtime checks)-------------------");
#else
    puts("--------------------------SPASS-START-----------------------------");
#endif

    puts("Input Problem:");
    clause_ListPrint(ProblemClauses);
    ana_Print(Flags, Precedence);
  }

  if (flag_GetFlagValue(Flags, flag_SORTS) != flag_SORTSOFF) {
    BOOL Strong;
    Strong = (flag_GetFlagValue(Flags, flag_SORTS) == flag_SORTSMONADICALL);
    for (Scan = ProblemClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
      clause_SetSortConstraint((CLAUSE)list_Car(Scan),Strong,Flags, Precedence);
  }

  if (flag_GetFlagValue(Flags, flag_RINPUT)) {
    clock_StartCounter(clock_REDUCTION);
    EmptyClauses = red_ReduceInput(Search, ProblemClauses);
    clock_StopAddPassedTime(clock_REDUCTION);
    if (list_Empty(EmptyClauses) && flag_GetFlagValue(Flags, flag_SATINPUT))
      EmptyClauses = red_SatInput(Search);
  }
  else {
    for (Scan=ProblemClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
  }
  Derivables = list_Nil();

  if (ana_SortRestrictions() ||
      (ana_UnsolvedSortRestrictions() &&
       flag_GetFlagValue(Flags,flag_SORTS) == flag_SORTSMONADICALL)) {
    if (flag_GetFlagValue(Flags, flag_RSST))
      prfs_SetStaticSortTheory(Search,sort_ApproxStaticSortTheory(prfs_UsableClauses(Search),Flags,Precedence));
    prfs_SetDynamicSortTheory(Search,sort_TheoryCreate());
  }

  /* Fix literal order in clauses in the usable set.
     Since they are shared, we have to extract them from
     the sharing before fixing them. Afterwards, we have to
     insert them in the sharing again.
  */ 
  for (Scan = prfs_UsableClauses(Search);
       !list_Empty(Scan);
       Scan = list_Cdr(Scan)) {
    CLAUSE clause;
    clause = list_Car(Scan);
    clause_MakeUnshared(clause,prfs_UsableSharingIndex(Search));
    clause_FixLiteralOrder(clause, Flags, Precedence);
    clause_InsertIntoSharing(clause, prfs_UsableSharingIndex(Search),
			     prfs_Store(Search), prfs_Precedence(Search));
  }

  /* Calculate the frequency counts for the symbols in the usable set. */
  for (Scan = prfs_UsableClauses(Search);
       !list_Empty(Scan);
       Scan = list_Cdr(Scan)) {
    CLAUSE clause;
    clause = list_Car(Scan);

    clause_CountSymbols(clause);
  }

  /* Sort usable set. */
  prfs_SetUsableClauses(Search, 
			list_Sort(prfs_UsableClauses(Search), 
				  (BOOL (*) (void *, void *)) clause_CompareAbstractLEQ));

  if (flag_GetFlagValue(Flags, flag_SOS)) {
    Derivables = list_Copy(prfs_UsableClauses(Search));
    for (Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (!clause_GetFlag(list_Car(Scan), CONCLAUSE))
	prfs_MoveUsableWorkedOff(Search,list_Car(Scan));
    list_Delete(Derivables);
  }

  if (1) {   /* EXPERIMENTAL CODE: NO INFERENCES BETWEEN THEORY CLAUSES */
    Derivables = list_Copy(prfs_UsableClauses(Search));
    for (Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (clause_IsTheoryAxiom(list_Car(Scan)))
      /* if (clause_IsTheoryStandardAxiom(list_Car(Scan))) */
	prfs_MoveUsableWorkedOff(Search,list_Car(Scan));
    list_Delete(Derivables);
  }

  if (flag_GetFlagValue(Flags, flag_PPROBLEM)) {
    puts("\nProcessed Problem:");
    puts("\nWorked Off Clauses:");
    clause_ListPrint(prfs_WorkedOffClauses(Search));
    puts("\nUsable Clauses:");
    clause_ListPrint(prfs_UsableClauses(Search));
  }

  /* SZ > */
  /* export usable clauses to DP */
  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
    clock_StartCounter(clock_SMT);
    /* give usable clauses to SMT procedure */
    for (Scan=prfs_UsableClauses(Search); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      dpconnect_exportClause(Search, list_Car(Scan));
    }
    clock_StopAddPassedTime(clock_SMT);
  }
  /* < SZ */

  while ((list_Empty(EmptyClauses) || !prfs_SplitStackEmpty(Search)) && 
	 /* SZ > */ (*smt_sat == 0)  && /* < SZ */
	 prfs_Loops(Search) != 0 &&
	 ((*BoundApplied != -1) || !list_Empty(prfs_UsableClauses(Search))) &&
	 (flag_GetFlagValue(Flags,flag_TIMELIMIT) == flag_TIMELIMITUNLIMITED ||
	  flag_GetFlagValue(Flags,flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL)) 
         /* (UW) */
         ) {

    Derivables    = list_Nil();
    SplitClause   = (CLAUSE)NULL;
    *BoundApplied = -1;

    while ((list_Empty(EmptyClauses) || !prfs_SplitStackEmpty(Search)) && 
	   /* SZ > */ (*smt_sat == 0)  && /* < SZ */
	   prfs_Loops(Search) != 0 &&
	   (!list_Empty(prfs_UsableClauses(Search)) || !list_Empty(EmptyClauses)) &&
	   (flag_GetFlagValue(Flags,flag_TIMELIMIT) == flag_TIMELIMITUNLIMITED ||
	    flag_GetFlagValue(Flags,flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))
           /* (UW) */
           ) {

      /*        profile_introduceDelayByMicroSecs(1000000);  */
      
      /* SZ > */
      /* check whether we have a result from dpconnect */
      if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	clock_StartCounter(clock_SMT);

	if (dpconnect_hasResult(Search) == 1) {
	  
	  smt_result = dpconnect_getCurrentSMTResult(Search);
	  
	  if (!smt_result_seen(dpconnect_getCurrentSMTResult(Search))) {
	    if (top_letSMTContribute(Search, Flags, smt_result,
				     &EmptyClauses,
				     &HighestLevelEmptyClause,
				     smt_sat) != 0) {
	      misc_StartErrorReport();
	      misc_ErrorReport("\n In top_top_ProofSearch: ");
	      misc_ErrorReport("Error during message receiving/handling. ");
	      misc_FinishErrorReport();
	    }
	  }
	}
	clock_StopAddPassedTime(clock_SMT);
      }
      /* < SZ */

      if (!list_Empty(EmptyClauses)) {
	/* Backtracking */
	
	clock_StartCounter(clock_BACKTRACK);
	/* SZ > */
	Derivables = split_BacktrackAndInformDPC(Search, HighestLevelEmptyClause,
						 &SplitClause);
	/* < SZ */
	clock_StopAddPassedTime(clock_BACKTRACK);

	prfs_IncBacktrackedClauses(Search, list_Length(Derivables));
	
	if (prfs_SplitStackEmpty(Search))
	  Derivables = list_Nconc(EmptyClauses, Derivables);
	else {
	  for ( ; !list_Empty(EmptyClauses); EmptyClauses = list_Pop(EmptyClauses))
	    if (list_Car(EmptyClauses) != HighestLevelEmptyClause)
	      clause_Delete(list_Car(EmptyClauses));
	  prfs_InsertDocProofClause(Search, HighestLevelEmptyClause);
	  /* Keep HighestLevelEmptyClause for finding the terms required */
	  /* for the proof.                                              */
	  if (flag_GetFlagValue(Flags, flag_DOCPROOF))
	    UsedEmptyClauses = list_Cons(HighestLevelEmptyClause, UsedEmptyClauses);
	  if (flag_GetFlagValue(Flags, flag_DOCSPLIT)) {
	    printf("\n\t Split Backtracking level %d:",prfs_ValidLevel(Search));
	    if (flag_GetFlagValue(Flags, flag_DOCSPLIT) == 2) {
	      LIST Scon;
	      printf("\n\t Backtracked clauses:");
	      for(Scon=Derivables;!list_Empty(Scon);Scon=list_Cdr(Scon)) {
		fputs("\n\tBclause: ", stdout);
		clause_Print((CLAUSE)list_Car(Scon));
		fflush(stdout);
	      }
	      printf("\n\t End Backtracked clauses:");
	    }
	  }
	}
	EmptyClauses            = list_Nil();
	HighestLevelEmptyClause = (CLAUSE)NULL;

      }
      else {   /* no empty clause found */

#ifdef CHECK
	if (!prfs_Check(Search)) {
	  misc_StartErrorReport();
	  misc_ErrorReport("\n In top_ProofSearch: Illegal state of search in SPASS.\n");
	  misc_FinishErrorReport();
	}
	if (!ana_Equations())
	  red_CheckSplitSubsumptionCondition(Search);
#endif 

	if (flag_GetFlagValue(Flags, flag_FULLRED))
	  Derivables = top_FullReductionSelectGivenComputeDerivables(Search, &SplitClause, &Counter, /* SZ > */ &SplitHeuristicCounter /* < SZ */);
	else
	  Derivables = top_LazyReductionSelectGivenComputeDerivables(Search, &SplitClause, &Counter, /* SZ > */ &SplitHeuristicCounter /* < SZ */);

      }

      /* Print the derived clauses, if required */
      if (flag_GetFlagValue(Flags, flag_PDER))
	for (Scan=Derivables; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	  fputs("\nDerived: ", stdout); 
	  clause_Print(list_Car(Scan));
	}

      /* (UW) */
      if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	clock_StartCounter(clock_SMT);
        for (Scan=Derivables; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	  /* SZ > */

	  dpconnect_exportClause(Search, list_Car(Scan));
	  /* < SZ */
	}
	clock_StopAddPassedTime(clock_SMT);

      }

      /* Partition the derived clauses into empty and non-empty clauses */
      Derivables = split_ExtractEmptyClauses(Derivables, &EmptyClauses);

      /* Apply reduction rules */
      clock_StartCounter(clock_REDUCTION);
      if (flag_GetFlagValue(Flags, flag_FULLRED)) {
	EmptyClauses = 
	  list_Nconc(EmptyClauses,
		     red_CompleteReductionOnDerivedClauses(Search, Derivables,
							   red_ALL, ActBound,
							   BoundMode,
							   BoundApplied));
      } else {
	EmptyClauses =
	  list_Nconc(EmptyClauses,
		     red_CompleteReductionOnDerivedClauses(Search, Derivables,
							   red_WORKEDOFF,
							   ActBound, BoundMode,
							   BoundApplied));
      }
      clock_StopAddPassedTime(clock_REDUCTION);
      
      if (!list_Empty(EmptyClauses)) {
	HighestLevelEmptyClause = split_SmallestSplitLevelClause(EmptyClauses);
	if (flag_GetFlagValue(Flags, flag_PEMPTYCLAUSE)) {
	  fputs("\nEmptyClause: ", stdout);
	  clause_Print(HighestLevelEmptyClause);
	}
      }
      prfs_DecLoops(Search);

      /* open branches: 
	           list_empty(EmptyClauses)  (no contradiction so far)
		      and
          	   list_empty(Usables)       (input saturated)

	    => SPASS would leave the loop and saturate its input
	 we have to ask the SMT procedure whether there is a theory inconsistency 
	 in case there is an SMT procedure, SPASS is automatically supposed 
	 to backtrack to the right level (if there is something to backtrack to), otherwise
	 in "top_ProofSearch" we know that <EmptyClauses != {} > and thus the formula
	 was refuted */
      
      if (list_Empty(prfs_UsableClauses(Search)) && list_Empty(EmptyClauses)) {
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  /* send an urgent check request */
	  clock_StartCounter(clock_SMT);
	  
#ifdef SMT_DEBUG
	  smt_debug_start();
	  printf("Performing an urgent check (since we have an open branch)...\n");
	  smt_debug_end();
#endif
#ifdef SMT_LOG
	  dpconnect_logMessage(Search, 1, "Performing an urgent check.\n");
#endif
	  dpconnect_performCheckUrgent(Search);
	  
	  /* wait, until urgent check delivered a result; 
	     perhaps this loop is interrupted by a signal from outside */
#ifdef SMT_DEBUG
	  smt_debug_start();
	  printf("Waiting for a result...\n");
	  smt_debug_end();
#endif

	  while(dpconnect_updateState(Search) != 1) {
	    if (flag_GetFlagValue(Flags,flag_TIMELIMIT) != flag_TIMELIMITUNLIMITED &&
		flag_GetFlagValue(Flags,flag_TIMELIMIT) <= clock_GetSeconds(clock_OVERALL)) {
	      break;
	    }
	  }
	  
	  /* handle the result */
	  if (dpconnect_hasResult(Search) == 1) {
	    smt_result = dpconnect_getCurrentSMTResult(Search);
	    
	    /* in case checking was interrupted, do not take an old 
	       (already seen) result */
	    if (!smt_result_seen(dpconnect_getCurrentSMTResult(Search)))
	      if (top_letSMTContribute(Search, Flags, smt_result,
				       &EmptyClauses,
				       &HighestLevelEmptyClause,
				       smt_sat) != 0) {
		misc_StartErrorReport();
		misc_ErrorReport("\n In top_top_ProofSearch: ");
		misc_ErrorReport("Error during message receiving/handling. ");
		misc_FinishErrorReport();
	      }
	  }
	  clock_StopAddPassedTime(clock_SMT);
	}
      }
    }

    /* outermost while loop */

    if (ActBound != flag_BOUNDSTARTUNLIMITED &&
	BoundMode != flag_BOUNDMODEUNLIMITED) {
      BoundLoops--;
      if (BoundLoops == flag_BOUNDLOOPSMIN)
	ActBound = flag_BOUNDSTARTUNLIMITED;
      else
	ActBound = *BoundApplied;
      if (*BoundApplied != -1) {
	prfs_SwapIndexes(Search);
	if  (flag_GetFlagValue(Flags,flag_PBINC))
	  printf("\n\n\t ---- New Clause %s Bound: %2d ----\n",
		 (BoundMode==flag_BOUNDMODERESTRICTEDBYDEPTH) ? "Term Depth" : "Weight",ActBound);
      }
    }
  }

  /* > SZ < moved here: */
  prfs_SetEmptyClauses(Search, EmptyClauses);
  prfs_SetUsedEmptyClauses(Search, UsedEmptyClauses);

  return Search;
}


static void top_Flotter(int argc, const char* argv[], LIST InputClauses)
/**************************************************************
  INPUT:  
  RETURNS: Nothing.
  EFFECT:  
***************************************************************/
{
  FILE *Output;
  char *description;
  const char *creator = "\n\tCNF generated by FLOTTER " misc_VERSION " *}";
  int  size;
  int  creator_size;

  if (argc < opts_Indicator()+2)
    Output = stdout;
  else
    Output = misc_OpenFile(argv[opts_Indicator()+1],"w");

  creator_size = (int)strlen(creator);
  size        = (int)strlen(dfg_ProblemDescription()) + creator_size;
  description = (char*)memory_Malloc(sizeof(char)*size);
  strncpy(description,dfg_ProblemDescription(),size-creator_size-3);
  strcpy(description+size-creator_size-3, creator);

  
  clause_FPrintCnfDFGProblem(Output, dfg_ProblemName(), dfg_ProblemAuthor(), 
			     dfg_ProblemStatusString(), description, InputClauses);
  
  fputs("\nFLOTTER needed\t", stdout);
  clock_PrintTime(clock_INPUT);
  puts(" for the input.");
  fputs("\t\t", stdout);
  clock_PrintTime(clock_CNF);
  fputs(" for the  CNF translation.", stdout);
  

  if (Output != stdout)
    misc_CloseFile(Output,argv[opts_Indicator()+1]);
  memory_Free(description, sizeof(char)*size);
}

static BOOL top_CalledFlotter(const char* Call)
{
  int length;
  length = strlen(Call);
  return string_Equal((Call + (length > 7 ? length - 7 : 0)), "FLOTTER");
}



/**************************************************************/
/* Main Function                                              */
/**************************************************************/

int main(int argc, const char* argv[])
{
  LIST              InputClauses,Scan,Axioms,Conjectures, Sorts, QueryClauses, 
    LabelClauses, QueryPair, ProblemClauses, Labellist, Sortlabellist, 
    Symblist, UserPrecedence;
  PROOFSEARCH       Search, FlotterSearch;
  /* <InputFlags> are the flags from the problem file and the command line. */
  FLAGSTORE         InputFlags, Flags;
  /* <InputPrecedence> is the precedence after reading the problem file. */
  PRECEDENCE        InputPrecedence, Precedence;
  FILE*             InputStream; 
  HASH              TermLabelToClauselist, ClauseToTermLabellist;
  top_SEARCHRESULT  Result;
  /* SZ > */
  int               smt_sat;
  /* < SZ */
  
  Search = (PROOFSEARCH)NULL;
  
#ifdef SPASS_SIGNALS
  top_PROOFSEARCH = &Search;
  signal(SIGINT, top_SigHandler);
  signal(SIGTERM, top_SigHandler);
  signal(SIGSEGV, top_SigHandler);
  signal(SIGBUS, top_SigHandler);
#endif

  clock_Init();
  clock_StartCounter(clock_OVERALL);
  memory_Init(memory__UNLIMITED);
  atexit(memory_FreeAllMem);
  /* SZ > */
  /* Initialize the SPASS connection to the GNU MP library
     do this before the initialization of the "symbol" module */
  gmp_Init();
  /* < SZ */
  symbol_Init(TRUE);
  stack_Init();
  hash_Init();
  sort_Init();
  term_Init();

  InputPrecedence = symbol_CreatePrecedence();
  fol_Init(InputPrecedence);
  cont_Init();
  unify_Init();
  flag_Init();
  subs_Init();
  clause_Init();
  red_Init();
  ren_Init();
  dp_Init();
  opts_Init();
  ana_Init();
  cc_Init();

  /* SZ > */
  clock_InitCounter(clock_SMT);
#ifdef SMT_DEBUG
  clock_InitCounter(clock_SMTCREATEINPUT);
  clock_InitCounter(clock_SMTGETRESULT);
#endif
  /* < SZ */

  /* Build proof search object to store definitions in */
  Search      = prfs_Create();
  InputFlags  = flag_CreateStore();

  /* declare all options */
  opts_DeclareSPASSFlagsAsOptions();
  top_RemoveFileOptId = opts_Declare("rf", opts_NOARGTYPE);

  if (!opts_Read(argc, argv)) 
    return EXIT_FAILURE;

   /* Check whether flag_STDIN is set in the command line */
  flag_InitStoreByDefaults(InputFlags);
  opts_SetFlags(InputFlags);

  if (argc < opts_Indicator()+1 && !flag_GetFlagValue(InputFlags,flag_STDIN)) {
    /* No input file, no stdin input */
    printf("\n\t          %s %s",argv[0],misc_VERSION);
    if (top_CalledFlotter(argv[0]) ||
	flag_GetFlagValue(InputFlags, flag_FLOTTER))
      puts("\n\t       Usage: FLOTTER [options] [<input-file>] [<output-file>]\n");
    else
      puts("\n\t       Usage: SPASS [options] [<input-file>] \n");
    puts("Possible options:\n");
    opts_PrintSPASSNames(); 
    return EXIT_FAILURE;
  }
  FlotterSearch = NULL;

  Axioms = Conjectures = Sorts = Labellist = Sortlabellist = UserPrecedence = list_Nil();
  
  if (flag_GetFlagValue(InputFlags, flag_STDIN)) {
    top_InputFile = (char*)NULL;
    InputStream   = stdin;
  } else {
    top_InputFile = argv[opts_Indicator()];
    InputStream = misc_OpenFile(top_InputFile, "r");
  }

  clock_StartCounter(clock_INPUT);
  flag_CleanStore(InputFlags);  /* Mark all flags as unset */
  
  /* Now add flags from file to InputFlags and set precedence */
 
  InputClauses = dfg_DFGParser(InputStream, InputFlags, InputPrecedence, &Axioms,
			       &Conjectures, &Sorts, &UserPrecedence); 
  
  /* Add/overwrite with command line flags */
  opts_SetFlags(InputFlags);
  Flags      = prfs_Store(Search);
  Precedence = prfs_Precedence(Search);
  /* The Flags were initialized with the default flag values.  */
  /* This values are now overwritten by command line flags and flags */
  /* from the input file. */
  flag_TransferSetFlags(InputFlags, Flags);
  /* From now on the input flags are not changed! */

  /* Transfer input precedence to search object */
  symbol_TransferPrecedence(InputPrecedence, Precedence);

  /* Complain about missing input clauses/formulae when in */
  /* non-interactive mode */
  if (!flag_GetFlagValue(Flags, flag_INTERACTIVE) && list_Empty(Axioms) &&
      list_Empty(Conjectures) && list_Empty(InputClauses)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n No formulae and clauses found in input file!\n");
    misc_FinishUserErrorReport();
  }

  cnf_Init(Flags);  /* Depends on Strong Skolemization Flag */

  /* check whether a DFG_FRAC was used but the theory is integer arithmetic;
     this cannot be done directly in the grammar rules since the 
     "DFG_SETFLAG"s are parsed after the clauses */
  if ((flag_GetFlagValue(Flags, flag_DPCTHEORY) != flag_DPCTRAT) &&
      (flag_GetFlagValue(Flags, flag_DPCTHEORY) != flag_DPCTREAL) &&
      top_FIRSTFRAC > 0) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: Fraction used, but theory is not rational or real arithmetic.\n", top_FIRSTFRAC);
    misc_FinishUserErrorReport();
  }

  /* DocProof is required for interactive mode */
  if (flag_GetFlagValue(Flags, flag_INTERACTIVE)) {
    flag_SetFlagValue(Flags, flag_DOCPROOF, flag_DOCPROOFON);
  }

  if (flag_GetFlagValue(Flags, flag_DOCPROOF))
    prfs_AddDocProofSharingIndex(Search);

  /* Create necessary hasharrays */
  if (flag_GetFlagValue(Flags, flag_DOCPROOF)) {
    TermLabelToClauselist = hsh_Create();
    ClauseToTermLabellist = hsh_Create();
  }
  else {
    TermLabelToClauselist = NULL;
    ClauseToTermLabellist = NULL;
  }
  
  /* Build conjecture formula and negate it: Conjectures are taken disjunctively!!*/
  for (Scan = Conjectures; !list_Empty(Scan); Scan = list_Cdr(Scan))
    list_Rplacd(list_Car(Scan), (LIST)term_Create(fol_Not(), 
						  list_List(list_PairSecond(list_Car(Scan)))));

  clock_StopPassedTime(clock_INPUT);

  if (top_InputFile) {
    misc_CloseFile(InputStream,top_InputFile);
    if (opts_IsSet(top_RemoveFileOptId))
      remove(top_InputFile);
  }

  clock_StartCounter(clock_CNF);

  if (list_Empty(InputClauses)) {
    NAT Termcount;

    Termcount  = 0;
    
    /* Create labels for formulae without them */
    for (Scan = Axioms; !list_Empty(Scan); Scan = list_Cdr(Scan), Termcount++) {
      if (list_PairFirst(list_Car(Scan)) == NULL) {
	char L[100];
	char* Label;
	sprintf(L, "axiom%d", Termcount);
	Label = string_StringCopy(L);
	list_Rplaca(list_Car(Scan), Label);
	if (flag_GetFlagValue(Flags, flag_DOCPROOF) &&
	    flag_GetFlagValue(Flags, flag_PLABELS)) {
	  printf("\nAdded label %s for axiom \n", Label);
	  fol_PrettyPrintDFG((TERM) list_PairSecond(list_Car(Scan)));
	}
      }
    }
    Termcount  = 0;
    for (Scan = Sorts; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      char L[100];
      char* Label;
      sprintf(L, "declaration%d", Termcount);
      Label = string_StringCopy(L);
      list_Rplaca(list_Car(Scan), Label);
      if (flag_GetFlagValue(Flags, flag_DOCPROOF) &&
	  flag_GetFlagValue(Flags, flag_PLABELS)) {
	printf("\nAdded label %s for declaration \n", Label);
	fol_PrettyPrintDFG((TERM) list_PairSecond(list_Car(Scan)));
      }
      Sortlabellist = list_Cons(Label, Sortlabellist);
    }
    Axioms = list_Nconc(Axioms, Sorts);


    if (flag_GetFlagValue(Flags, flag_APPLYDEFS) != flag_APPLYDEFSOFF) {
      def_ExtractDefsFromTermlist(Search, Axioms, Conjectures); 
      Conjectures = def_ApplyDefinitionToTermList(prfs_Definitions(Search),
						  Conjectures, Flags,
						  Precedence);
    }

    /* We must keep the list of symbols because they can't be freed in cnf_Flotter */
    Symblist = list_Nil();

    /* Axioms is list of pairs, conjectures is list of terms */
    /* A ProofSearch object is only returned and the symbols kept in Symblist
       if flag_INTERACTIVE is set */
    FlotterSearch = cnf_Flotter(Axioms,Conjectures,&InputClauses, &Labellist,
				TermLabelToClauselist, ClauseToTermLabellist,
				Flags, Precedence, &Symblist);
    
    InputClauses = clause_ListSortWeighed(InputClauses);

    if (flag_GetFlagValue(Flags, flag_DPCTHEORY) >= flag_DPCTINT
        && flag_GetFlagValue(Flags, flag_ARITHABS) != flag_ARITHABSTRACTIONOFF) {
      for (Scan = InputClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
        CLAUSE C;
        C = list_Car(Scan);
        list_Rplaca(Scan,clause_AbstractTheorySubterms(C, 2, Flags, Precedence));
        clause_Delete(C);
      }
    }

    clause_SetCounter(1);
    for (Scan = InputClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      clause_NewNumber(list_Car(Scan));
    }
  }
  else {
    dfg_DeleteFormulaPairList(Axioms);
    dfg_DeleteFormulaPairList(Sorts);
    dfg_DeleteFormulaPairList(Conjectures);
    if (flag_GetFlagValue(Flags, flag_APPLYDEFS) != flag_APPLYDEFSOFF) {
      /* Extract list of definitions */
      def_ExtractDefsFromClauselist(Search, InputClauses);
      def_FlattenDefinitionsDestructive(Search);
      for (Scan=prfs_Definitions(Search); !list_Empty(Scan); Scan=list_Cdr(Scan))
	InputClauses = def_ApplyDefToClauselist(Search, (DEF) list_Car(Scan),
						InputClauses, TRUE);
    }
  }

  clock_StopPassedTime(clock_CNF);

  if (top_CalledFlotter(argv[0]) || flag_GetFlagValue(Flags, flag_FLOTTER)) {
    top_Flotter(argc,argv,InputClauses);
    flag_SetFlagValue(Flags, flag_TIMELIMIT,   0);       /* Exit No Output */
    flag_SetFlagValue(Flags, flag_INTERACTIVE, flag_INTERACTIVEOFF);
    flag_SetFlagValue(Flags, flag_PPROBLEM,    flag_PPROBLEMOFF);
  }

  memory_Restrict(flag_GetFlagValue(Flags, flag_MEMORY));

  do {
    LIST deflist;
    int  BoundApplied;
    ProblemClauses = list_Nil();
    LabelClauses   = list_Nil();
    Result         = top_RESOURCE;

    if (flag_GetFlagValue(Flags, flag_INTERACTIVE)) {
      QueryPair = ia_GetNextRequest(InputStream, Flags);
      /* A pair (<formula,labellist>) */
      /* Get list of clauses derivable from formulae with labels in labellist */
      if (list_Empty(QueryPair)) {
	break;
      }
      for (Scan=list_PairSecond(QueryPair);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	LIST l;
	l = hsh_GetWithCompareFunc(TermLabelToClauselist, list_Car(Scan),
				   (BOOL (*)(POINTER, POINTER)) cnf_LabelEqual,
				   (unsigned long (*)(POINTER)) hsh_StringHashKey);

	l = list_PointerDeleteDuplicates(list_Copy(l));
	LabelClauses = list_Nconc(l, LabelClauses);
      }
      /* Get list of clauses derivable from sorts */
      for (Scan=Sortlabellist; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	LIST l;
	l = hsh_GetWithCompareFunc(TermLabelToClauselist, list_Car(Scan),
				   (BOOL (*)(POINTER, POINTER)) cnf_LabelEqual,
				   (unsigned long (*)(POINTER)) hsh_StringHashKey);

	l = list_PointerDeleteDuplicates(list_Copy(l));
	LabelClauses = list_Nconc(l, LabelClauses);
      }	

      /* For labelclauses copies are introduced */
      /* So an update to the clause to term mapping is necessary */
      for (Scan=LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	CLAUSE C;
	LIST l;
	C = (CLAUSE) list_Car(Scan);
	l = list_Copy(hsh_Get(ClauseToTermLabellist, C));
	l = cnf_DeleteDuplicateLabelsFromList(l);
	list_Rplaca(Scan, clause_Copy(C));
	hsh_PutList(ClauseToTermLabellist, list_Car(Scan), l);
      }
      QueryClauses   = cnf_QueryFlotter(FlotterSearch, list_PairFirst(QueryPair),
					&Symblist);
      ProblemClauses = list_Nconc(QueryClauses, LabelClauses);

      for (Scan=list_PairSecond(QueryPair); !list_Empty(Scan); Scan= list_Cdr(Scan))
	string_StringFree(list_Car(Scan)); /* Free the label strings */
      list_Delete(list_PairSecond(QueryPair));
      list_PairFree(QueryPair);
      clock_InitCounter(clock_OVERALL);
      clock_StartCounter(clock_OVERALL);
    }
    else {
      ProblemClauses = InputClauses;
      InputClauses   = list_Nil();
    }

    prfs_SetSplitCounter(Search,flag_GetFlagValue(Flags, flag_SPLITS));
    prfs_SetLoops(Search,flag_GetFlagValue(Flags, flag_LOOPS));
    prfs_SetBacktrackedClauses(Search, 0);
    BoundApplied = -1;

    /* (UW) */
    if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
      /* SZ > */
      clock_StartCounter(clock_SMT);
      /* initialize dpconnect */
      dpconnect_Init(Search);
      clock_StopAddPassedTime(clock_SMT);
      /* < SZ */
    }

    /* SZ > */
    if (flag_GetFlagValue(Flags,flag_DPCTHEORY) >= flag_DPCTINT &&
        flag_GetFlagValue(Flags,flag_THAXIOMS) == flag_ON)
	/* add fixed standard axioms to <ProblemClauses>, before the analyzation 
	   routines are called and the weights are recomputed */
	top_AddStandardAxioms(Search, &ProblemClauses, flag_GetFlagValue(Flags,flag_DPCTHEORY));
    /* < SZ */
    
    Search = top_ProofSearch(Search, ProblemClauses, InputFlags, UserPrecedence, 
			     &BoundApplied /* SZ > */, &smt_sat /* < SZ */);

    /* interpret result of "top_ProofSearch" */
    if ((flag_GetFlagValue(Flags, flag_TIMELIMIT) == flag_TIMELIMITUNLIMITED ||
	 flag_GetFlagValue(Flags, flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL)) &&
	prfs_Loops(Search) != 0 &&
	(BoundApplied == -1 || !list_Empty(prfs_EmptyClauses(Search)))) {
   
      if (list_Empty(prfs_EmptyClauses(Search))) {
        Result = top_COMPLETION;
        /* UW */
        /* We do not make use of a satisfiability result produced by the
           SMT procedure:
           For the free theory, completion already indicates satisfiability,
           so the SMT result is irrelevant.
	   For arithmetic theories, completion does not indicate
           satisfiability, but since the translation to SMT may be lossy,
	   satisfiability of the translated formulas does not imply
	   satisfiability of the original formulas either. */
        /*
        if (smt_sat != 1) {
  	  Result = top_COMPLETION;
        } else {
  	  Result = top_SMT_SAT;
	}
        */
      } else {
	/* SZ > */
	if (prfs_GetClosedBranchesTheoryInconsistency(Search) == prfs_GetClosedBranches(Search)
	    && (prfs_GetClosedBranchesTheoryInconsistency(Search) > 0))
	  Result = top_SMT;
	else if (prfs_GetClosedBranchesTheoryInconsistency(Search) == 0)
	  Result = top_PROOF;
	else
	  Result = top_SPASS_SMT;
	/* < SZ */
      }
    }
    
    if (flag_GetFlagValue(Flags, flag_TIMELIMIT) != 0) {
      /* Stop SPASS immediately */
      /*
      printf("\nSPASS %s ", misc_VERSION);
      if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
      } */

      /* fputs("\nSPASS beiseite: ", stdout); */
      
      switch (Result) {
      case top_RESOURCE:
	/* (UW) */
	if (prfs_Loops(Search) != 0) {
          printf("\n\n%% SZS status Timeout for %s\n",
                 (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	  printf("\nSPASS %s ", misc_VERSION);
	  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	    printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	  }
          fputs("\nSPASS beiseite: Ran out of time.", stdout);
	} else {
          printf("\n\n%% SZS status GaveUp for %s\n",
                 (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	  printf("\nSPASS %s ", misc_VERSION);
	  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	    printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	  }
	  fputs("\nSPASS beiseite: Maximal number of loops exceeded.", stdout);
	}
	break;
      case top_PROOF:
        printf("\n\n%% SZS status Theorem for %s\n",
               (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	printf("\nSPASS %s ", misc_VERSION);
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	}
	fputs("\nSPASS beiseite: Proof found by SPASS.", stdout);
	break;
      /* SZ > */
      case top_SMT:
        printf("\n\n%% SZS status Theorem for %s\n",
               (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	printf("\nSPASS %s ", misc_VERSION);
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	}
	fputs("\nSPASS beiseite: Proof found by SMT.", stdout);
	break;
      case top_SPASS_SMT:
        printf("\n\n%% SZS status Theorem for %s\n",
               (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	printf("\nSPASS %s ", misc_VERSION);
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	}
	fputs("\nSPASS beiseite: Proof found by SPASS and SMT.", stdout);
	break;
      /* < SZ */
      case top_SMT_SAT:
        /* this branch is currently unused */
        printf("\n\n%% SZS status CounterSatisfiable for %s\n",
               (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	printf("\nSPASS %s ", misc_VERSION);
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	}
	fputs("\nSPASS beiseite: Satisfiable (found by SMT procedure).", stdout);
	break;
      default: /* top_COMPLETION */
        if (flag_GetFlagValue(Flags,flag_DPCTHEORY) >= flag_DPCTINT) {
	  printf("\n\n%% SZS status Incomplete for %s\n",
		 (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	  printf("\nSPASS %s ", misc_VERSION);
	  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	    printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	  }
	  fputs("\nSPASS beiseite: Unknown (SPASS (with standard axioms): completion", stdout);
	  
	  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_OFF) {
	    fputs(", SMT procedure: unknown", stdout);
	  }
	  fputs(").", stdout);
	} else {
	  printf("\n\n%% SZS status CounterSatisfiable for %s\n",
		 (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	  printf("\nSPASS %s ", misc_VERSION);
	  if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	    printf(" in combination with %s.", smt_interface_getName((prfs_DpConnect(Search))->smt_interface));
	  }
	  fputs("\nSPASS beiseite: Satisfiable.", stdout);
	}
      }
      printf("\nProblem: %s ",
	     (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
      if (flag_GetFlagValue(Flags, flag_PSTATISTIC)) {
	clock_StopPassedTime(clock_OVERALL);
	printf("\nSPASS derived %d clauses,", prfs_DerivedClauses(Search));
	printf(" backtracked %d clauses", prfs_BacktrackedClauses(Search));
	printf(" and kept %d clauses.", prfs_KeptClauses(Search));
	printf("\nSPASS backtracked %d times (%d times due to theory inconsistency).",
	       prfs_GetClosedBranches(Search),
	       prfs_GetClosedBranchesTheoryInconsistency(Search));
	printf("\nSPASS allocated %lu KBytes.", memory_DemandedBytes()/1024);
	fputs("\nSPASS spent\t", stdout);
	clock_PrintTime(clock_OVERALL);
	fputs(" on the problem.\n\t\t", stdout);
	clock_PrintTime(clock_INPUT);
	fputs(" for the input.\n\t\t", stdout);
	clock_PrintTime(clock_CNF);
	fputs(" for the FLOTTER CNF translation.\n\t\t", stdout);
	clock_PrintTime(clock_INFERENCE);
	fputs(" for inferences.\n\t\t", stdout);
        clock_PrintTime(clock_BACKTRACK);
	fputs(" for the backtracking.\n\t\t", stdout);
	clock_PrintTime(clock_REDUCTION);
	fputs(" for the reduction.", stdout);
	/* SZ > */
	if (flag_GetFlagValue(Flags,flag_DPCPROVER) != flag_DPCOFF) {
	  fputs("\n\t\t", stdout);
	  clock_PrintTime(clock_SMT);
	  fputs(" for interacting with the SMT procedure.\n\t\t", stdout);
#ifdef SMT_DEBUG
	  fputs("\n\t\t", stdout);
	  clock_PrintTime(clock_SMTCREATEINPUT);
	  fputs(" for providing input to the SMT procedure.", stdout);

	  fputs("\n\t\t", stdout);
	  clock_PrintTime(clock_SMTGETRESULT);
	  fputs(" for {looking for,processing} the SMT result.\n", stdout);
#endif
	}
	/* < SZ */
      }
      if (Result != top_PROOF &&
	  flag_GetFlagValue(Flags, flag_FPMODEL) != flag_FPMODELOFF) {
	FILE *Output;
	char name[100];
	const char * creator = "{*SPASS " misc_VERSION " *}";
	BOOL PrintPotProductive;

	strcpy(name, (top_InputFile != (char*)NULL ? top_InputFile : "SPASS"));
	if (Result == top_COMPLETION)
	  strcat(name, ".model");
	else
	  strcat(name, ".clauses");
	Output = misc_OpenFile(name,"w");
	PrintPotProductive = (flag_GetFlagValue(Flags, flag_FPMODEL) ==
			      flag_FPMODELPOTENTIALLYPRODUCTIVECLAUSES);
	if (Result == top_COMPLETION)
	  clause_FPrintCnfFormulasDFGProblem(Output, PrintPotProductive,
					     "{*Completion_by_SPASS*}",
					     creator, "satisfiable",
					     dfg_ProblemDescription(),
					     prfs_WorkedOffClauses(Search),
					     list_Nil(), Flags, Precedence);
	else
	  clause_FPrintCnfFormulasDFGProblem(Output, PrintPotProductive,
					     "{*Clauses_generated_by_SPASS*}",
					     creator, "unknown",
					     dfg_ProblemDescription(),
					     prfs_WorkedOffClauses(Search),
					     prfs_UsableClauses(Search), Flags,
					     Precedence);
	misc_CloseFile(Output, name);
	if (Result == top_COMPLETION)
	  printf("\nCompletion printed to: %s\n", name);
	else
	  printf("\nClauses printed to: %s\n", name);
      }
    
      if (flag_GetFlagValue(Flags, flag_DOCPROOF) && Result != top_RESOURCE) {
	if (Result == top_COMPLETION) {
	  /* the completion does not induce a model (like it does for pure EUF) 
	     because the arithmetic knowledge is incomplete; this is reported
	     explicitely */
          /* UW */
          /* The following statements seems to be too confusing for users */
	  /* puts("\n\nA *potential* model is given by the following saturation."); */
          printf("\n\n%% SZS output start ListOfCNF for %s",
                (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	  puts("\n\nThe saturated set of worked-off clauses is :");
	  clause_ListPrint(prfs_WorkedOffClauses(Search));
          printf("\n\n%% SZS output end ListOfCNF for %s",
                (top_InputFile != (char*)NULL ? top_InputFile : "stdin"));
	} else {
	  LIST UsedClauses, UsedTerms;
	  /* if (!top_InputFile)
	    top_InputFile = "SPASS"; */
	  UsedClauses = dp_PrintProof(Search, prfs_EmptyClauses(Search),
				      top_InputFile,
                                      top_InputFile != (char*)NULL ? top_InputFile : "SPASS");
	  /* Find terms required for proof */
	  UsedTerms = list_Nil();

	  for (Scan = UsedClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
	    if (clause_IsFromInput((CLAUSE) list_Car(Scan))) {
	      LIST L;
	      L = hsh_Get(ClauseToTermLabellist, list_Car(Scan));
	      L = list_Copy(L);
	      L = cnf_DeleteDuplicateLabelsFromList(L);
	      UsedTerms = list_Nconc(UsedTerms, L);
	    }
	  list_Delete(UsedClauses);
	  UsedTerms = cnf_DeleteDuplicateLabelsFromList(UsedTerms);
	  fputs("\nFormulae used in the proof :", stdout);
	  for (Scan = UsedTerms; !list_Empty(Scan); Scan = list_Cdr(Scan)) 
	    if (!(strncmp((char*) list_Car(Scan), "_SORT_", 6) == 0))
	      printf(" %s", (char*) list_Car(Scan));
	  putchar('\n');
	  list_Delete(UsedTerms);
	}
      }
    }
    
    /* Delete mapping for the clause copies of the labelclauses */
    for (Scan = LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
      hsh_DelItem(ClauseToTermLabellist, list_Car(Scan));
    
    list_Delete(ProblemClauses);
 
    fflush(stdout);

    /* Keep definitions */
    deflist = prfs_Definitions(Search);
    prfs_SetDefinitions(Search, list_Nil());
    prfs_Clean(Search);
    prfs_SetDefinitions(Search, deflist);

    symbol_TransferPrecedence(InputPrecedence, Precedence);
    if (flag_GetFlagValue(Flags, flag_PPROBLEM))
      fputs("\n--------------------------SPASS-STOP------------------------------", stdout);
  } while (flag_GetFlagValue(Flags, flag_INTERACTIVE) &&
	   (flag_GetFlagValue(Flags, flag_TIMELIMIT) != 0));

  for (Scan = InputClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
    clause_OrientAndReInit(list_Car(Scan), Flags, Precedence);

  /* Cleanup Flotter data structures */
  if (flag_GetFlagValue(Flags, flag_INTERACTIVE)) {
    if (flag_GetFlagValue(Flags, flag_DOCPROOF))
      list_Delete(Symblist);
    else 
      symbol_DeleteSymbolList(Symblist);
    /*  symbol_ResetSkolemIndex(); */
    if (FlotterSearch != NULL) 
      prfs_Delete(FlotterSearch);
  }
  if (flag_GetFlagValue(Flags, flag_PFLAGS)) {
    putchar('\n');
    flag_Print(Flags);
  }
  if (flag_GetFlagValue(Flags, flag_DOCPROOF)) {
    hsh_Delete(TermLabelToClauselist);
    hsh_Delete(ClauseToTermLabellist);
  }
  for (Scan = Labellist; !list_Empty(Scan); Scan = list_Cdr(Scan))
    string_StringFree(list_Car(Scan));
  list_Delete(Labellist);
  list_Delete(Sortlabellist);
  list_Delete(UserPrecedence);

  cnf_Free(Flags);

  prfs_Delete(Search);
  clause_DeleteClauseList(InputClauses);
  flag_DeleteStore(InputFlags);
  symbol_DeletePrecedence(InputPrecedence);

  cc_Free();
  ana_Free();
  sort_Free();
  unify_Free();
  cont_Free();
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
  opts_Free();

  /* SZ > */
  /* Finishing the SPASS connection to the GNU MP library
     do this after the finishing of the "symbol" module */
  gmp_Free();
  /* < SZ */
  
#ifdef CHECK
  memory_Print();
  memory_PrintLeaks();
#endif
  putchar('\n');
  return 0;
}
